from nengo.networks.assoc_mem import AssociativeMemory as AssocMem
from nengo.spa.module import Module


class AssociativeMemory(Module):
    """Associative memory module.

    See :doc:`examples/spa/associative_memory` for an introduction and
    examples.

    Parameters
    ----------
    input_vocab: Vocabulary
        The vocabulary to match.
    output_vocab: Vocabulary, optional (Default: None)
        The vocabulary to be produced for each match. If
        None, the associative memory will act like an autoassociative memory
        (cleanup memory).
    input_keys : list, optional (Default: None)
        A list of strings that correspond to the input vectors.
    output_keys : list, optional (Default: None)
        A list of strings that correspond to the output vectors.
    default_output_key: str, optional (Default: None)
        The semantic pointer string to be produced if the input value matches
        none of vectors in the input vector list.
    threshold: float, optional (Default: 0.3)
        The association activation threshold.
    inhibitable: bool, optional (Default: False)
        Flag to indicate if the entire associative memory module is
        inhibitable (i.e., the entire module can be inhibited).
    wta_output: bool, optional (Default: False)
        Flag to indicate if output of the associative memory should contain
        more than one vector. If True, only one vector's output will be
        produced; i.e. produce a winner-take-all (WTA) output.
        If False, combinations of vectors will be produced.
    wta_inhibit_scale: float, optional (Default: 3.0)
        Scaling factor on the winner-take-all (WTA) inhibitory connections.
    wta_synapse: float, optional (Default: 0.005)
        Synapse to use for the winner-take-all (wta) inhibitory connections.
    threshold_output: bool, optional (Default: False)
        Adds a threholded output if True.
    label : str, optional (Default: None)
        A name for the ensemble. Used for debugging and visualization.
    seed : int, optional (Default: None)
        The seed used for random number generation.
    add_to_container : bool, optional (Default: None)
        Determines if this Network will be added to the current container.
        If None, will be true if currently within a Network.
    """

    def __init__(self, input_vocab, output_vocab=None,  # noqa: C901
                 input_keys=None, output_keys=None,
                 default_output_key=None, threshold=0.3,
                 inhibitable=False, wta_output=False,
                 wta_inhibit_scale=3.0, wta_synapse=0.005,
                 threshold_output=False, label=None, seed=None,
                 add_to_container=None):
        super(AssociativeMemory, self).__init__(label, seed, add_to_container)

        if input_keys is None:
            input_keys = input_vocab.keys
            input_vectors = input_vocab.vectors
        else:
            input_vectors = [input_vocab.parse(key).v for key in input_keys]

        # If output vocabulary is not specified, use input vocabulary
        # (i.e autoassociative memory)
        if output_vocab is None:
            output_vocab = input_vocab
            output_vectors = input_vectors
        else:
            if output_keys is None:
                output_keys = input_keys

            output_vectors = [output_vocab.parse(key).v for key in output_keys]

        if default_output_key is None:
            default_output_vector = None
        else:
            default_output_vector = output_vocab.parse(default_output_key).v

        # Create nengo network
        with self:
            self.am = AssocMem(input_vectors=input_vectors,
                               output_vectors=output_vectors,
                               threshold=threshold,
                               inhibitable=inhibitable,
                               label=label, seed=seed,
                               add_to_container=add_to_container)

            if default_output_vector is not None:
                self.am.add_default_output_vector(default_output_vector)

            if wta_output:
                self.am.add_wta_network(wta_inhibit_scale, wta_synapse)

            if threshold_output:
                self.am.add_threshold_to_outputs()

            self.input = self.am.input
            self.output = self.am.output

            if inhibitable:
                self.inhibit = self.am.inhibit

            self.utilities = self.am.utilities
            if threshold_output:
                self.thresholded_utilities = self.am.thresholded_utilities

        self.inputs = dict(default=(self.input, input_vocab))
        self.outputs = dict(default=(self.output, output_vocab))



"""Functions for easy interactions with IPython and IPython notebooks.

NotebookRunner is modified from runipy.
This modified code is included under the terms of its license:

Copyright (c) 2013, Paul Butler
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from __future__ import absolute_import

import io
import os
import platform
import time
import uuid

import numpy as np

try:
    import IPython
    from IPython import get_ipython
    from IPython.display import HTML

    if IPython.version_info[0] <= 3:
        from IPython.config import Config
        from IPython.nbconvert import HTMLExporter, PythonExporter
    else:
        from traitlets.config import Config
        from nbconvert import HTMLExporter, PythonExporter

    # nbformat.current deprecated in IPython 3.0
    if IPython.version_info[0] <= 2:
        # pylint: disable=ungrouped-imports
        from IPython.nbformat import current
        from IPython.nbformat.current import write as write_nb
        from IPython.nbformat.current import NotebookNode

        def read_nb(fp):
            return current.read(fp, 'json')
    else:
        if IPython.version_info[0] == 3:
            # pylint: disable=ungrouped-imports
            from IPython import nbformat
            from IPython.nbformat import write as write_nb
            from IPython.nbformat import NotebookNode
        else:
            import nbformat
            from nbformat import write as write_nb
            from nbformat import NotebookNode

        def read_nb(fp):
            # Have to load as version 4 or running notebook fails
            return nbformat.read(fp, 4)
except ImportError:
    def get_ipython():
        return None
assert get_ipython


def check_ipy_version(min_version):
    try:
        import IPython
        return IPython.version_info >= min_version
    except ImportError:
        return False


def has_ipynb_widgets():
    """Determines whether IPython widgets are available.

    Returns
    -------
    bool
        ``True`` if IPython widgets are available, otherwise ``False``.
    """
    try:
        if IPython.version_info[0] <= 3:
            from IPython.html import widgets as ipywidgets
            from IPython.utils import traitlets
        else:
            import ipywidgets
            import traitlets
        assert ipywidgets
        assert traitlets
    except ImportError:
        return False
    else:
        return True


def hide_input():
    """Hide the input of the Jupyter notebook input block this is executed in.

    Returns a link to toggle the visibility of the input block.
    """
    uuid = np.random.randint(np.iinfo(np.int32).max)

    script = """
        <a id="%(uuid)s" href="javascript:toggle_input_%(uuid)s()"
          >Show Input</a>

        <script type="text/javascript">
        var toggle_input_%(uuid)s;
        (function() {
            if (typeof jQuery == 'undefined') {
                // no jQuery
                var link_%(uuid)s = document.getElementById("%(uuid)s");
                var cell = link_%(uuid)s;
                while (cell.className.split(' ')[0] != "cell"
                       && cell.className.split(' ')[0] != "nboutput") {
                    cell = cell.parentNode;
                }
                var input_%(uuid)s;
                if (cell.className.split(' ')[0] == "cell") {
                    for (var i = 0; i < cell.children.length; i++) {
                        if (cell.children[i].className.split(' ')[0]
                            == "input") {
                            input_%(uuid)s = cell.children[i];
                        }
                    }
                } else {
                    input_%(uuid)s = cell.previousElementSibling;
                }
                input_%(uuid)s.style.display = "none"; // hide

                toggle_input_%(uuid)s = function() {
                    if (input_%(uuid)s.style.display == "none") {
                        input_%(uuid)s.style.display = ""; // show
                        link_%(uuid)s.innerHTML = "Hide Input";
                    } else {
                        input_%(uuid)s.style.display = "none"; // hide
                        link_%(uuid)s.innerHTML = "Show Input";
                    }
                }

            } else {
                // jQuery
                var link_%(uuid)s = $("a[id='%(uuid)s']");
                var cell_%(uuid)s = link_%(uuid)s.parents("div.cell:first");
                if (cell_%(uuid)s.length == 0) {
                    cell_%(uuid)s = link_%(uuid)s.parents(
                        "div.nboutput:first");
                }
                var input_%(uuid)s = cell_%(uuid)s.children("div.input");
                if (input_%(uuid)s.length == 0) {
                    input_%(uuid)s = cell_%(uuid)s.prev("div.nbinput");
                }
                input_%(uuid)s.hide();

                toggle_input_%(uuid)s = function() {
                    if (input_%(uuid)s.is(':hidden')) {
                        input_%(uuid)s.slideDown();
                        link_%(uuid)s[0].innerHTML = "Hide Input";
                    } else {
                        input_%(uuid)s.slideUp();
                        link_%(uuid)s[0].innerHTML = "Show Input";
                    }
                }
            }
        }());
        </script>
    """ % dict(uuid=uuid)

    return HTML(script)


def load_notebook(nb_path):
    with io.open(nb_path, 'r', encoding='utf-8') as f:
        nb = read_nb(f)
    return nb


def export_py(nb, dest_path=None):
    """Convert notebook to Python script.

    Optionally saves script to dest_path.
    """
    exporter = PythonExporter()
    body, resources = exporter.from_notebook_node(nb)

    # Remove all lines with get_ipython
    while u"get_ipython()" in body:
        ind0 = body.find(u"get_ipython()")
        ind1 = body.find(u"\n", ind0)
        body = body[:ind0] + body[(ind1 + 1):]

    if u"plt" in body:
        body += u"\nplt.show()\n"

    if dest_path is not None:
        with io.open(dest_path, 'w', encoding='utf-8') as f:
            f.write(body)
    return body


def export_html(nb, dest_path=None, image_dir=None, image_rel_dir=None):
    """Convert notebook to HTML.

    Optionally saves HTML to dest_path.
    """
    c = Config({'ExtractOutputPreprocessor': {'enabled': True}})

    exporter = HTMLExporter(template_file='full', config=c)
    output, resources = exporter.from_notebook_node(nb)
    header = output.split('<head>', 1)[1].split('</head>', 1)[0]
    body = output.split('<body>', 1)[1].split('</body>', 1)[0]

    # Monkeypatch CSS
    header = header.replace('<style', '<style scoped="scoped"')
    header = header.replace(
        'body {\n  overflow: visible;\n  padding: 8px;\n}\n', '')
    header = header.replace("code,pre{", "code{")

    # Filter out styles that conflict with the sphinx theme.
    bad_anywhere = ['navbar',
                    'body{',
                    'alert{',
                    'uneditable-input{',
                    'collapse{']
    bad_anywhere.extend(['h%s{' % (i+1) for i in range(6)])

    bad_beginning = ['pre{', 'p{margin']

    header_lines = [x for x in header.split('\n')
                    if (not any(x.startswith(s) for s in bad_beginning)
                        and not any(s in x for s in bad_anywhere))]
    header = '\n'.join(header_lines)

    # Concatenate raw html lines
    lines = ['<div class="ipynotebook">']
    lines.append(header)
    lines.append(body)
    lines.append('</div>')
    html_out = '\n'.join(lines)

    if image_dir is not None and image_rel_dir is not None:
        html_out = export_images(resources, image_dir, image_rel_dir, html_out)

    if dest_path is not None:
        with io.open(dest_path, 'w', encoding='utf-8') as f:
            f.write(html_out)
    return html_out


def export_images(resources, image_dir, image_rel_dir, html_out):
    my_uuid = uuid.uuid4().hex

    for output in resources['outputs']:
        fname = "%s%s" % (my_uuid, output)
        new_path = os.path.join(image_dir, fname)
        new_rel_path = os.path.join(image_rel_dir, fname)
        html_out = html_out.replace(output, new_rel_path)
        with open(new_path, 'wb') as f:
            f.write(resources['outputs'][output])
    return html_out


def export_evaluated(nb, dest_path=None, skip_exceptions=False):
    """Convert notebook to an evaluated notebook.

    Optionally saves the notebook to dest_path.
    """
    nb_runner = NotebookRunner(nb)
    nb_runner.run_notebook(skip_exceptions=skip_exceptions)

    if dest_path is not None:
        with io.open(dest_path, 'w', encoding='utf-8') as f:
            if IPython.version_info[0] <= 2:
                write_nb(nb_runner.nb, f, 'ipynb')
            else:
                write_nb(nb_runner.nb, f)
    return nb_runner.nb


class NotebookRunner(object):
    # The kernel communicates with mime-types while the notebook
    # uses short labels for different cell types. We'll use this to
    # map from kernel types to notebook format types.

    MIME_MAP = {
        'image/jpeg': 'jpeg',
        'image/png': 'png',
        'text/plain': 'text',
        'text/html': 'html',
        'text/latex': 'latex',
        'application/javascript': 'html',
        'image/svg+xml': 'svg',
    }

    def __init__(self, nb, working_dir=None):
        from IPython.kernel import KernelManager

        self.km = KernelManager()

        cwd = os.getcwd()
        if working_dir is not None:
            os.chdir(working_dir)
        self.km.start_kernel()
        os.chdir(cwd)

        if platform.system() == 'Darwin':
            # There is sometimes a race condition where the first
            # execute command hits the kernel before it's ready.
            # It appears to happen only on Darwin (Mac OS) and an
            # easy (but clumsy) way to mitigate it is to sleep
            # for a second.
            time.sleep(1)

        self.kc = self.km.client()
        self.kc.start_channels()
        self.shell = self.kc.shell_channel
        self.iopub = self.kc.iopub_channel
        self.nb = nb

    def __del__(self):
        self.kc.stop_channels()
        self.km.shutdown_kernel(now=True)

    def run_cell(self, cell):  # noqa: C901
        """Run a notebook cell and update the output of that cell in-place."""
        self.shell.execute(cell.input)
        reply = self.shell.get_msg()
        status = reply['content']['status']
        if status == 'error':
            traceback_text = ("Cell raised uncaught exception: \n"
                              "\n".join(reply['content']['traceback']))

        outs = []
        while True:
            msg = self.iopub.get_msg(timeout=1)
            msg_type = msg['msg_type']
            content = msg['content']

            if msg_type == 'status' and content['execution_state'] == 'idle':
                break

            # IPython 3.0.0-dev writes pyerr/pyout in the notebook format
            # but uses error/execute_result in the message spec. This does
            # the translation needed for tests to pass with IPython 3.0.0-dev
            notebook3_format_conversions = {
                'error': 'pyerr',
                'execute_result': 'pyout',
            }
            msg_type = notebook3_format_conversions.get(msg_type, msg_type)

            out = NotebookNode(output_type=msg_type)

            if 'execution_count' in content:
                cell['prompt_number'] = content['execution_count']
                out.prompt_number = content['execution_count']

            if msg_type in ('status', 'pyin', 'execute_input',
                            'comm_open', 'comm_msg'):
                continue
            elif msg_type == 'stream':
                out.stream = content['name']
                out.text = content['data']
            elif msg_type in ('display_data', 'pyout'):
                for mime, data in content['data'].items():
                    try:
                        attr = self.MIME_MAP[mime]
                    except KeyError:
                        raise NotImplementedError(
                            'unhandled mime type: %s' % mime)
                    if "widgets/js/widget" in data:
                        continue
                    setattr(out, attr, data)
            elif msg_type == 'pyerr':
                out.ename = content['ename']
                out.evalue = content['evalue']
                out.traceback = content['traceback']
            elif msg_type == 'clear_output':
                outs = []
                continue
            else:
                raise NotImplementedError(
                    'unhandled iopub message: %s' % msg_type)
            outs.append(out)
        cell['outputs'] = outs

        if status == 'error':
            raise Exception(traceback_text)

    def iter_code_cells(self):
        """Iterate over the notebook cells containing code."""
        for ws in self.nb.worksheets:
            for cell in ws.cells:
                if cell.cell_type == 'code':
                    yield cell

    def run_notebook(self, skip_exceptions=False, progress_callback=None):
        """Runs all notebook cells in order and updates outputs in-place.

        If ``skip_exceptions`` is True, then if exceptions occur in a cell, the
        subsequent cells are run (by default, the notebook execution stops).
        """
        for i, cell in enumerate(self.iter_code_cells()):
            try:
                self.run_cell(cell)
            except Exception:
                if not skip_exceptions:
                    raise
            if progress_callback is not None:
                progress_callback(i)

    def count_code_cells(self):
        """Return the number of code cells in the notebook."""
        return sum(1 for _ in self.iter_code_cells())



from . import builder
from . import cache
from . import compat
from . import connection
from . import ensemble
from . import filter_design
from . import functions
from . import graphs
from . import ipython
from . import logging
from . import magic
from . import nco
from . import network
from . import neurons
from . import numpy
from . import paths
from . import probe
from . import simulator
from . import stdlib



import nengo
from nengo import Network as BaseNetwork

from nengolib import Network
from nengolib.stats import ScatteredHypersphere


def test_network():
    with Network():
        x = nengo.Ensemble(100, 1)
        assert isinstance(x.eval_points, ScatteredHypersphere)
        assert isinstance(x.encoders, ScatteredHypersphere)

    with BaseNetwork():
        x = nengo.Ensemble(100, 1)
        assert not isinstance(x.eval_points, ScatteredHypersphere)
        assert not isinstance(x.encoders, ScatteredHypersphere)


def test_ensemble_array():
    with Network():
        ea = nengo.networks.EnsembleArray(100, 2)
        for ens in ea.ea_ensembles:
            assert isinstance(ens.eval_points, ScatteredHypersphere)
            assert isinstance(ens.encoders, ScatteredHypersphere)



"""
Extra functions to extend the capabilities of Numpy.
"""
from __future__ import absolute_import

import numpy as np

from .compat import PY2, is_integer, is_iterable
from ..exceptions import ValidationError

maxint = np.iinfo(np.int32).max


def compare(a, b):
    return 0 if a == b else 1 if a > b else -1 if a < b else None


def as_shape(x, min_dim=0):
    """Return a tuple if ``x`` is iterable or ``(x,)`` if ``x`` is integer."""
    if is_iterable(x):
        shape = tuple(x)
    elif is_integer(x):
        shape = (x,)
    else:
        raise ValueError("%r cannot be safely converted to a shape" % x)

    if len(shape) < min_dim:
        shape = tuple([1] * (min_dim - len(shape))) + shape

    return shape


def broadcast_shape(shape, length):
    """Pad a shape with ones following standard Numpy broadcasting."""
    n = len(shape)
    if n < length:
        return tuple([1] * (length - n) + list(shape))
    else:
        return shape


def array(x, dims=None, min_dims=0, readonly=False, **kwargs):
    y = np.array(x, **kwargs)
    dims = max(min_dims, y.ndim) if dims is None else dims

    if y.ndim < dims:
        shape = np.ones(dims, dtype='int')
        shape[:y.ndim] = y.shape
        y.shape = shape
    elif y.ndim > dims:
        raise ValidationError("Input cannot be cast to array with "
                              "%d dimensions" % dims, attr='dims')

    if readonly:
        y.flags.writeable = False

    return y


def array_hash(a, n=100):
    """Simple fast array hash function.

    For arrays with size larger than ``n``, pick ``n`` elements at random
    to hash. This strategy should work well for dense matrices, but for
    sparse ones it is more likely to give hash collisions.
    """
    if not isinstance(a, np.ndarray):
        return hash(a)

    if a.size < n:
        # hash all elements
        v = a.view()
        v.setflags(write=False)
        return hash(v.data if PY2 else v.data.tobytes())
    else:
        # pick random elements to hash
        rng = np.random.RandomState(a.size)
        inds = [rng.randint(0, a.shape[i], size=n) for i in range(a.ndim)]
        v = a[inds]
        v.setflags(write=False)
        return hash(v.data if PY2 else v.data.tobytes())


def array_offset(x):
    """Get offset of array data from base data in bytes."""
    if x.base is None:
        return 0

    base_start = x.base.__array_interface__['data'][0]
    start = x.__array_interface__['data'][0]
    return start - base_start


def expm(A, n_factors=None, normalize=False):
    """Simple matrix exponential to replace Scipy's matrix exponential

    This just uses a recursive (factored) version of the Taylor series,
    and is not as good as Scipy (which uses Pade approximants). The hard
    part with this implementation is choosing the length of the Taylor
    series. A longer series is generally needed for a matrix with a larger
    eigenvalues, but I'm not exactly sure how these relate. I'm using
    a heuristic based on the matrix norm, since this is kind-of related
    to the size of eigenvalues, though for larger norms the function
    becomes inaccurate no matter the length of the series.

    This function is mostly intended for use in `filter_design`, where
    the matrices should be small, both in dimensions and norm.
    """
    if A.ndim != 2 or A.shape[0] != A.shape[1]:
        raise ValidationError("'A' must be a square matrix", attr='A')

    a = np.linalg.norm(A)
    if normalize:
        a = int(a)
        A = A / float(a)

    if n_factors is None:
        n_factors = 20 if normalize else min(max(int(a), 20), 1000)

    Y = np.zeros_like(A)
    for i in range(n_factors, 0, -1):
        Y = np.dot(A / float(i), Y)
        np.fill_diagonal(Y, Y.diagonal() + 1)  # add identity matrix

    return np.linalg.matrix_power(Y, a) if normalize else Y


def norm(x, axis=None, keepdims=False):
    """Euclidean norm

    Parameters
    ----------
    x : array_like
        Array to compute the norm over.
    axis : None or int or tuple of ints, optional
        Axis or axes to sum across. `None` sums all axes. See `np.sum`.
    keepdims : bool, optional
        If True, the reduced axes are left in the result. See `np.sum` in
        newer versions of Numpy (>= 1.7).
    """
    return np.sqrt(np.sum(x**2, axis=axis, keepdims=keepdims))


def meshgrid_nd(*args):
    args = [np.asarray(a) for a in args]
    s = len(args) * (1,)
    return np.broadcast_arrays(*(
        a.reshape(s[:i] + (-1,) + s[i + 1:]) for i, a in enumerate(args)))


def rms(x, axis=None, keepdims=False):
    """Root-mean-square amplitude

    Parameters
    ----------
    x : array_like
        Array to compute RMS amplitude over.
    axis : None or int or tuple of ints, optional
        Axis or axes to sum across. `None` sums all axes. See `np.sum`.
    keepdims : bool, optional
        If True, the reduced axes are left in the result. See `np.sum` in
        newer versions of Numpy (>= 1.7).
    """
    return np.sqrt(np.mean(x**2, axis=axis, keepdims=keepdims))


def rmse(x, y, axis=None, keepdims=False):
    """Root-mean-square error amplitude

    Equivalent to rms(x - y, axis=axis, keepdims=keepdims).

    Parameters
    ----------
    x, y : array_like
        Arrays to compute RMS amplitude over.
    axis : None or int or tuple of ints, optional
        Axis or axes to sum across. `None` sums all axes. See `np.sum`.
    keepdims : bool, optional
        If True, the reduced axes are left in the result. See `np.sum` in
        newer versions of Numpy (>= 1.7).
    """
    return rms(x - y, axis=axis, keepdims=keepdims)


if hasattr(np.fft, 'rfftfreq'):
    rfftfreq = np.fft.rfftfreq
else:
    def rfftfreq(n, d=1.0):
        return np.abs(np.fft.fftfreq(n=n, d=d)[:n // 2 + 1])



from copy import deepcopy
import warnings

from nengo.config import Config
from nengo.connection import Connection
from nengo.ensemble import Ensemble
from nengo.exceptions import (
    ConfigError, NetworkContextError, NotAddedToNetworkWarning, ReadonlyError)
from nengo.node import Node
from nengo.params import IntParam, StringParam
from nengo.probe import Probe
from nengo.utils.compat import iteritems
from nengo.utils.threading import ThreadLocalStack


class Network(object):
    """A network contains ensembles, nodes, connections, and other networks.

    A network is primarily used for grouping together related
    objects and connections for visualization purposes.
    However, you can also use networks as a nice way to reuse
    network creation code.

    To group together related objects that you do not need to reuse,
    you can create a new ``Network`` and add objects in a ``with`` block.
    For example::

        network = nengo.Network()
        with network:
            with nengo.Network(label="Vision"):
                v1 = nengo.Ensemble(nengo.LIF(100), dimensions=2)
            with nengo.Network(label="Motor"):
                sma = nengo.Ensemble(nengo.LIF(100), dimensions=2)
            nengo.Connection(v1, sma)

    To reuse a group of related objects, you can create a new subclass
    of ``Network``, and add objects in the ``__init__`` method.
    For example::

        class OcularDominance(nengo.Network):
            def __init__(self):
                self.column = nengo.Ensemble(nengo.LIF(100), dimensions=2)

        network = nengo.Network()
        with network:
            left_eye = OcularDominance()
            right_eye = OcularDominance()
            nengo.Connection(left_eye.column, right_eye.column)

    Parameters
    ----------
    label : str, optional (Default: None)
        Name of the network.
    seed : int, optional (Default: None)
        Random number seed that will be fed to the random number generator.
        Setting the seed makes the network's build process deterministic.
    add_to_container : bool, optional (Default: None)
        Determines if this network will be added to the current container.
        If None, this network will be added to the network at the top of the
        ``Network.context`` stack unless the stack is empty.

    Attributes
    ----------
    connections : list
        `.Connection` instances in this network.
    ensembles : list
        `.Ensemble` instances in this network.
    label : str
        Name of this network.
    networks : list
        `.Network` instances in this network.
    nodes : list
        `.Node` instances in this network.
    probes : list
        `.Probe` instances in this network.
    seed : int
        Random seed used by this network.
    """

    context = ThreadLocalStack(maxsize=100)  # static stack of Network objects

    label = StringParam('label', optional=True, readonly=False)
    seed = IntParam('seed', optional=True, readonly=False)

    def __init__(self, label=None, seed=None, add_to_container=None):
        self.label = label
        self.seed = seed
        self._config = self.default_config()

        self.objects = {
            Ensemble: [], Node: [], Connection: [], Network: [], Probe: [],
        }
        self.ensembles = self.objects[Ensemble]
        self.nodes = self.objects[Node]
        self.connections = self.objects[Connection]
        self.networks = self.objects[Network]
        self.probes = self.objects[Probe]

        # By default, we want to add to the current context, unless there is
        # no context; i.e., we're creating a top-level network.
        if add_to_container is None:
            add_to_container = len(Network.context) > 0

        if add_to_container:
            Network.add(self)

    @staticmethod
    def add(obj):
        """Add the passed object to ``Network.context``."""
        if len(Network.context) == 0:
            raise NetworkContextError(
                "'%s' must either be created inside a ``with network:`` "
                "block, or set add_to_container=False in the object's "
                "constructor." % obj)
        network = Network.context[-1]
        if not isinstance(network, Network):
            raise NetworkContextError(
                "Current context (%s) is not a network" % network)
        for cls in type(obj).__mro__:
            if cls in network.objects:
                network.objects[cls].append(obj)
                break
        else:
            raise NetworkContextError("Objects of type %r cannot be added to "
                                      "networks." % type(obj).__name__)

    @staticmethod
    def default_config():
        """Constructs a `~.Config` object for setting defaults."""
        return Config(Connection, Ensemble, Node, Probe)

    def _all_objects(self, object_type):
        """Returns a list of all objects of the specified type."""
        # Make a copy of this network's list
        objects = list(self.objects[object_type])
        for subnet in self.networks:
            objects.extend(subnet._all_objects(object_type))
        return objects

    @property
    def all_objects(self):
        """(list) All objects in this network and its subnetworks."""
        objects = []
        for object_type in self.objects:
            objects.extend(self._all_objects(object_type))
        return objects

    @property
    def all_ensembles(self):
        """(list) All ensembles in this network and its subnetworks."""
        return self._all_objects(Ensemble)

    @property
    def all_nodes(self):
        """(list) All nodes in this network and its subnetworks."""
        return self._all_objects(Node)

    @property
    def all_networks(self):
        """(list) All networks in this network and its subnetworks."""
        return self._all_objects(Network)

    @property
    def all_connections(self):
        """(list) All connections in this network and its subnetworks."""
        return self._all_objects(Connection)

    @property
    def all_probes(self):
        """(list) All probes in this network and its subnetworks."""
        return self._all_objects(Probe)

    @property
    def config(self):
        """(`.Config`) Configuration for this network."""
        return self._config

    @config.setter
    def config(self, dummy):
        raise ReadonlyError(attr='config', obj=self)

    @property
    def n_neurons(self):
        """(int) Number of neurons in this network, including subnetworks."""
        return sum(ens.n_neurons for ens in self.all_ensembles)

    def __contains__(self, obj):
        return type(obj) in self.objects and obj in self.objects[type(obj)]

    def __enter__(self):
        Network.context.append(self)
        self._config.__enter__()
        return self

    def __exit__(self, dummy_exc_type, dummy_exc_value, dummy_tb):
        if len(Network.context) == 0:
            raise NetworkContextError(
                "Network.context in bad state; was empty when "
                "exiting from a 'with' block.")

        config = Config.context[-1]
        if config is not self._config:
            raise ConfigError("Config.context in bad state; was expecting "
                              "current context to be '%s' but instead got "
                              "'%s'." % (self._config, config))

        network = Network.context.pop()
        if network is not self:
            raise NetworkContextError(
                "Network.context in bad state; was expecting current context "
                "to be '%s' but instead got '%s'." % (self, network))

        self._config.__exit__(dummy_exc_type, dummy_exc_value, dummy_tb)

    def __getstate__(self):
        state = self.__dict__.copy()
        state['label'] = self.label
        state['seed'] = self.seed
        return state

    def __setstate__(self, state):
        for k, v in iteritems(state):
            setattr(self, k, v)
        if len(Network.context) > 0:
            warnings.warn(NotAddedToNetworkWarning(self))

    def __str__(self):
        return "<%s %s>" % (
            type(self).__name__,
            '"%s"' % self.label if self.label is not None else
            "(unlabeled) at 0x%x" % id(self))

    def __repr__(self):
        return "<%s %s %s>" % (
            type(self).__name__,
            '"%s"' % self.label if self.label is not None else "(unlabeled)",
            "at 0x%x" % id(self))

    def copy(self, add_to_container=None):
        with warnings.catch_warnings():
            # We warn when copying since we can't change add_to_container.
            # However, we deal with it here, so we ignore the warning.
            warnings.simplefilter('ignore', category=NotAddedToNetworkWarning)
            c = deepcopy(self)
        if add_to_container is None:
            add_to_container = len(Network.context) > 0
        if add_to_container:
            Network.add(c)
        return c



from __future__ import absolute_import

from collections import OrderedDict
import warnings

import numpy as np

from nengo.exceptions import ValidationError
from nengo.utils.compat import is_number, iteritems


def function_name(func):
    """Returns the name of a function.

    Unlike accessing ``func.__name__``, this function is robust to the
    different types of objects that can be considered a function in Nengo.

    Parameters
    ----------
    func : callable or array_like
        Object used as function argument.

    Returns
    -------
    str
        Name of function object.
    """
    return getattr(func, "__name__", func.__class__.__name__)


def piecewise(data):
    """Create a piecewise constant function from a dictionary.

    .. deprecated:: 2.5.1
       Use `nengo.processes.Piecewise` instead.

    Given an input of data={0: 0, 0.5: 1, 0.75: -1, 1: 0} this will generate a
    function that returns 0 up until t=0.5, then outputs a 1 until t=0.75,
    then a -1 until t=1, and then returns 0 after that. This is meant as a
    shortcut for::

        def function(t):
            if t < 0.5:
                return 0
            elif t < 0.75
                return 1
            elif t < 1:
                return -1
            else:
                return 0

    The keys in the dictionary must be times (floats or ints). The values in
    the data dictionary can be floats, lists, or functions that return
    floats or lists. All lists must be of the same length.

    For times before the first specified time, it will default to zero (of
    the correct length). This means the above example can be simplified to::

        piecewise({0.5: 1, 0.75: -1, 1: 0})

    Parameters
    ----------
    data : dict
        The values to change to. Keys are the beginning time for the value.
        Values can be int, float, list, or functions that return those.

    Returns
    -------
    function:
        A function that takes a variable t and returns the corresponding
        value from the dictionary.

    Examples
    --------

      >>> func = piecewise({0.5: 1, 0.75: -1, 1: 0})
      >>> func(0.2)
      [0]
      >>> func(0.58)
      [1]

      >>> func = piecewise({0.5: [1, 0], 0.75: [0, 1]})
      >>> func(0.2)
      [0,0]
      >>> func(0.58)
      [1,0]
      >>> func(100)
      [0,1]

      >>> import math
      >>> func = piecewise({0: math.sin, 0.5: math.cos})
      >>> func(0.499)
      [0.47854771647582706]
      >>> func(0.5)
      [0.8775825618903728]

    """

    warnings.warn("The `piecewise` function is deprecated. Use the Piecewise "
                  "process instead.", DeprecationWarning)

    # first, sort the data (to simplify finding the right element
    # when calling the function)
    output_length = None  # the dimensionality of the returned values
    for time in data:
        if not is_number(time):
            raise ValidationError("Keys must be times (floats or ints), not %r"
                                  % type(time).__name__, attr='data')

        # figure out the length of this item
        if callable(data[time]):
            length = np.asarray(data[time](0.0)).size
        else:
            data[time] = np.asarray(data[time])
            length = data[time].size

        # make sure this is the same length as previous items
        if length != output_length and output_length is not None:
            raise ValidationError("time %g has %d items instead of %d" %
                                  (time, length, output_length), attr='data')
        output_length = length

    # make a default output of 0 when t before what was passed
    data[np.finfo(float).min] = np.zeros(output_length)
    ordered_data = OrderedDict(sorted(iteritems(data)))

    # build the function to return
    def piecewise_function(t, data=ordered_data):
        # get the t we'll use for output
        for time in (time for time in data if time <= t):
            out_t = time

        # if it's a function, call it
        if callable(data[out_t]):
            return np.asarray(data[out_t](t))
        return data[out_t]
    return piecewise_function


class HilbertCurve(object):
    """Hilbert curve function.

    Pre-calculates the Hilbert space filling curve with a given number
    of iterations. The curve will lie in the square delimited by the
    points (0, 0) and (1, 1).

    Arguments
    ---------
    n : int
        Iterations.
    """
    # Implementation based on
    # https://en.wikipedia.org/w/index.php?title=Hilbert_curve&oldid=633637210

    def __init__(self, n):
        self.n = n
        self.n_corners = (2 ** n) ** 2
        self.corners = np.zeros((self.n_corners, 2))
        self.steps = np.arange(self.n_corners)

        steps = np.arange(self.n_corners)
        for s in 2 ** np.arange(n):
            r = np.empty_like(self.corners, dtype='int')
            r[:, 0] = 1 & (steps // 2)
            r[:, 1] = 1 & (steps ^ r[:, 0])
            self._rot(s, r)
            self.corners += s * r
            steps //= 4

        self.corners /= (2 ** n) - 1

    def _rot(self, s, r):
        swap = r[:, 1] == 0
        flip = np.all(r == np.array([1, 0]), axis=1)

        self.corners[flip] = (s - 1 - self.corners[flip])
        self.corners[swap] = self.corners[swap, ::-1]

    def __call__(self, u):
        """Evaluate pre-calculated Hilbert curve.

        Arguments
        ---------
        u : ndarray (M,)
            Positions to evaluate on the curve in the range [0, 1].

        Returns
        -------
        ndarray (M, 2)
            Two-dimensional curve coordinates.
        """
        step = np.asarray(u * len(self.steps))
        return np.vstack((
            np.interp(step, self.steps, self.corners[:, 0]),
            np.interp(step, self.steps, self.corners[:, 1]))).T



import logging
import warnings
from random import random

import numpy as np

import nengo
from nengo.params import NumberParam, Default
from nengo.synapses import SynapseParam

from nengolib.network import Network
from nengolib.signal.realizers import Hankel
from nengolib.signal.system import LinearSystem
from nengolib.synapses.mapping import ss2sim

__all__ = ['LinearNetwork']


class LinearNetwork(Network):
    """Network implementing a linear time-invariant (LTI) system.

    This network implements the following linear state-space model:

    .. math::

       \\dot{{\\bf x}}(t) &= A{\\bf x}(t) + B{\\bf u}(t) \\\\
              {\\bf y}(t) &= C{\\bf x}(t) + D{\\bf u}(t)

    This works by first realizing a state-space representation from the
    given ``sys`` and ``realizer``, and then using :func:`.ss2sim` to apply
    a generalization of Principle 3 from the Neural Engineering Framework (NEF)
    to map the system onto the given ``synapse``. This yields a :attr:`.mapped`
    system whose state-space matrices give the transformation matrices for the
    resulting Nengo network.

    Parameters
    ----------
    sys : :data:`linear_system_like`
       Linear system representation.
    n_neurons_per_ensemble : ``integer``
       Number of neurons to use per ensemble (i.e., per dimension).
    synapse : :class:`nengo.synapses.Synapse`
        Recurrent synapse used to implement the dynamics, passed to
        :func:`.ss2sim`.
    dt : ``float`` or ``None``
        Simulation time-step (in seconds), passed to :func:`.ss2sim`.
        If ``None``, then this uses the continuous form of Principle 3
        (i.e., assuming a continuous-time synapse with negligible time-step).
        If provided, then ``sys`` will be discretized and the discrete
        form of Principle 3 will be applied. This should always be given
        for digital simulations.
    radii : ``float`` or ``array_like``, optional
        Radius of each dimension of the realized state-space.
        If a single ``float``, then it will be applied to each dimension.
        If ``array_like``, then its length must match :attr:`.size_state`.
        Defaults to ``1``.
    input_synapse : :class:`nengo.synapses.Synapse`, optional
        Input synapse connecting from :attr:`.input` node. Defaults to ``None``
        to discourage double filtering, but should typically match the
        ``synapse`` parameter.
    output_synapse : :class:`nengo.synapses.Synapse`, optional
        Output synapse connecting to :attr:`.output` node.
        Defaults to ``None``.
    realizer : :class:`.AbstractRealizer`, optional
        Method of obtaining a state-space realization of ``sys``.
        Defaults to :class:`.Hankel`.
    solver : :class:`nengo.solvers.Solver`, optional
        Solver used to decode the state.
        Defaults to :class:`nengo.solvers.LstsqL2` (with ``reg=.1``).
    label : str, optional (Default: None)
        Name of the network.
    seed : int, optional (Default: None)
        Random number seed that will be fed to the random number generator.
        Setting the seed makes the network's build process deterministic.
    add_to_container : bool, optional (Default: None)
        Determines if this network will be added to the current container.
        If None, this network will be added to the network at the top of the
        ``Network.context`` stack unless the stack is empty.
    **ens_kwargs : ``dictionary``, optional
        Additional keyword arguments are passed to the
        :class:`nengo.networks.EnsembleArray` that represents the
        :attr:`.state`.

    See Also
    --------
    :class:`.Network`
    :class:`.RollingWindow`
    :class:`.Hankel`
    :func:`.ss2sim`

    Notes
    -----
    By linearity, the ``input_synapse`` and the ``output_synapse`` are
    interchangeable with one another. However, this will modify the
    state-space (according to these same filters) which may impact the quality
    of representation.

    Examples
    --------
    >>> from nengolib.networks import LinearNetwork
    >>> from nengolib.synapses import Bandpass

    Implementing a 5 Hz :func:`.Bandpass` filter (i.e., a decaying 2D
    oscillator) using 1000 spiking LIF neurons:

    >>> import nengo
    >>> from nengolib import Network
    >>> from nengolib.signal import Balanced
    >>> with Network() as model:
    >>>     stim = nengo.Node(output=lambda t: 100*int(t < .01))
    >>>     sys = LinearNetwork(sys=Bandpass(freq=5, Q=10),
    >>>                         n_neurons_per_ensemble=500,
    >>>                         synapse=.1, dt=1e-3, realizer=Balanced())
    >>>     nengo.Connection(stim, sys.input, synapse=None)
    >>>     p = nengo.Probe(sys.state.output, synapse=.01)
    >>> with nengo.Simulator(model, dt=sys.dt) as sim:
    >>>     sim.run(1.)

    Note there are exactly 5 oscillations within 1 second, in response to a
    saturating impulse:

    >>> import matplotlib.pyplot as plt
    >>> plt.plot(*sim.data[p].T)
    >>> plt.xlabel("$x_1(t)$")
    >>> plt.ylabel("$x_2(t)$")
    >>> plt.axis('equal')
    >>> plt.xlim(-1, 1)
    >>> plt.ylim(-1, 1)
    >>> plt.show()
    """

    synapse = SynapseParam('synapse')
    input_synapse = SynapseParam('input_synapse', optional=True)
    output_synapse = SynapseParam('input_synapse', optional=True)
    dt = NumberParam('dt', low=0, low_open=True, optional=True)

    def __init__(self, sys, n_neurons_per_ensemble, synapse, dt, radii=1.0,
                 input_synapse=None, output_synapse=None,
                 realizer=Hankel(), solver=Default,
                 label=None, seed=None, add_to_container=None, **ens_kwargs):
        super(LinearNetwork, self).__init__(label, seed, add_to_container)

        # Parameter checking
        self.sys = LinearSystem(sys)
        self.n_neurons_per_ensemble = n_neurons_per_ensemble
        self.synapse = synapse
        self.dt = dt
        self.radii = radii
        self.input_synapse = input_synapse
        self.output_synapse = output_synapse
        self.realizer = realizer

        if solver is not Default:
            # https://github.com/nengo/nengo/issues/1044
            solver._hack = random()

        if len(self.sys) == 0:
            raise ValueError("system (%s) is zero order" % self.sys)

        if self.sys.has_passthrough and self.output_synapse is None:
            # the user shouldn't filter the output node themselves. an
            # output synapse should be given so we can do it before the
            # passthrough.
            warnings.warn("output_synapse should be given if the system has "
                          "a passthrough, otherwise filtering the output will "
                          "also filter the passthrough")

        if not self.sys.is_stable:
            # This means certain normalizers won't work, because the worst-case
            # output is now unbounded.
            warnings.warn("system (%s) is not exponentially stable" % self.sys)

        # Obtain state-space transformation and realization
        self._realizer_result = self.realizer(self.sys, self.radii)

        # Map the system onto the synapse
        self._mapped = ss2sim(self.realization, self.synapse, self.dt)

        with self:
            # Create internal Nengo objects
            self._input = nengo.Node(size_in=self.size_in, label="input")
            self._output = nengo.Node(size_in=self.size_out, label="output")

            x_input, x_output = self._make_core(solver, **ens_kwargs)

            # Connect everything up using (A, B, C, D)
            nengo.Connection(
                x_output, x_input, transform=self.A,
                synapse=self.synapse)
            nengo.Connection(
                self.input, x_input, transform=self.B,
                synapse=self.input_synapse)
            nengo.Connection(
                x_output, self.output, transform=self.C,
                synapse=self.output_synapse)

            if not np.allclose(self.D, 0):
                logging.info("Passthrough (%s) on LinearNetwork with sys=%s",
                             self.D, self.sys)
                nengo.Connection(
                    self.input, self.output, transform=self.D,
                    synapse=None)

    def _make_core(self, solver, **ens_kwargs):
        self._state = nengo.networks.EnsembleArray(
            self.n_neurons_per_ensemble, self.size_state,
            ens_dimensions=1, label="x", **ens_kwargs)

        if solver is not Default:
            # https://github.com/nengo/nengo/issues/1040
            self.state.add_output('output', function=None, solver=solver)

        return self.state.input, self.state.output

    @property
    def realizer_result(self):
        """The :class:`.RealizerResult` produced by ``realizer``."""
        return self._realizer_result

    @property
    def realization(self):
        """Realized :class:`.LinearSystem`."""
        return self.realizer_result.realization  # convenience

    @property
    def mapped(self):
        """Mapped :class:`.LinearSystem`."""
        return self._mapped

    @property
    def A(self):
        """``A`` state-space matrix of mapped :class:`.LinearSystem`."""
        return self.mapped.A

    @property
    def B(self):
        """``B`` state-space matrix of mapped :class:`.LinearSystem`."""
        return self.mapped.B

    @property
    def C(self):
        """``C`` state-space matrix of mapped :class:`.LinearSystem`."""
        return self.mapped.C

    @property
    def D(self):
        """``D`` state-space matrix of mapped :class:`.LinearSystem`."""
        return self.mapped.D

    @property
    def size_in(self):
        """Input dimensionality."""
        return self.mapped.size_in

    @property
    def size_state(self):
        """State dimensionality."""
        return len(self.mapped)

    @property
    def size_out(self):
        """Output dimensionality."""
        return self.mapped.size_out

    @property
    def input(self):
        """Nengo object representing the input ``u(t)`` to the system."""
        return self._input

    @property
    def state(self):
        """Nengo object representing the state ``x(t)`` of the system."""
        return self._state

    @property
    def output(self):
        """Nengo object representing the output ``y(t)`` of the system."""
        return self._output



import numpy as np

from nengo.base import Process
from nengo.dists import Choice, Distribution
from nengo.exceptions import ValidationError

from nengolib.signal.system import LinearSystem

__all__ = ['EvalPoints', 'Encoders']


class EvalPoints(Distribution):
    """Samples the output of a LinearSystem given some input process.

    This can be used to sample the evaluation points according to some
    filtered process. Used by :class:`.RollingWindow`.

    Parameters
    ----------
    sys : :data:`linear_system_like`
       Linear system representation.
    process : :class:`nengo.Process`
       Nengo process to simulate.
    n_steps : ``integer``, optional
       Number of steps to simulate the process. Defaults to ``10000``.
    dt : ``float``, optional
       Process and system simulation time-step.
       Defaults to ``process.default_dt``.
    **run_steps_kwargs : ``dictionary``, optional
       Additional keyword arguments for ``process.run_steps``.

    See Also
    --------
    :class:`.Encoders`
    :class:`.Callable`
    :class:`.RollingWindow`
    :class:`nengo.Ensemble`
    :class:`nengo.dists.Distribution`

    Notes
    -----
    For ideal sampling, the given ``process`` should be aperiodic across the
    interval of time specified by ``n_steps`` and ``dt``, and moreover
    the sampled ``num`` (number of evaluation points) should not
    exceed ``n_steps``.

    Examples
    --------
    >>> from nengolib.signal import EvalPoints

    Sampling from the state-space of an alpha synapse given band-limited
    white noise:

    >>> from nengolib import Alpha
    >>> from nengo.processes import WhiteSignal
    >>> eval_points = EvalPoints(Alpha(.5).X, WhiteSignal(10, high=20))

    >>> import matplotlib.pyplot as plt
    >>> from seaborn import jointplot
    >>> jointplot(*eval_points.sample(1000, 2).T, kind='kde')
    >>> plt.show()
    """

    def __init__(self, sys, process, n_steps=10000, dt=None,
                 **run_steps_kwargs):
        super(EvalPoints, self).__init__()
        self.sys = LinearSystem(sys)
        if not isinstance(process, Process):
            raise ValidationError(
                "process (%s) must be a Process" % (process,),
                attr='process', obj=self)
        self.process = process
        self.n_steps = n_steps
        if dt is None:
            dt = self.process.default_dt  # 0.001
        self.dt = dt
        self.run_steps_kwargs = run_steps_kwargs

    def __repr__(self):
        return ("%s(sys=%r, process=%r, n_steps=%r, dt=%r, **%r)" %
                (type(self).__name__, self.sys, self.process, self.n_steps,
                 self.dt, self.run_steps_kwargs))

    def _sample(self, d, rng):
        if self.sys.size_out != d:
            raise ValidationError(
                "sys.size_out (%d) must equal sample d (%s)" %
                (self.sys.size_out, d), attr='sys', obj=self)
        u = self.process.run_steps(
            self.n_steps, d=self.sys.size_in, dt=self.dt, rng=rng,
            **self.run_steps_kwargs)
        return self.sys.filt(u, dt=self.dt)

    def sample(self, num, d=1, rng=np.random):
        """Samples ``n`` points in ``d`` dimensions."""
        y = self._sample(d, rng)
        choices = rng.choice(len(y), size=num, replace=True)
        return y[choices]


class Encoders(EvalPoints):
    """Samples axis-aligned encoders from the maximum radii of a LinearSystem.

    Given some input process. This is useful when setting ``encoders`` in
    conjunction with ``normalize_encoders=False``, in order to set the
    *effective* radius of each dimension to its maximum absolute value.
    The evaluation points should still be sampled from within this radii.
    Paramters are the same as :class:`.EvalPoints`.
    Used by :class:`.RollingWindow`.

    Parameters
    ----------
    sys : :data:`linear_system_like`
       Linear system representation.
    process : :class:`nengo.Process`
       Nengo process to simulate.
    n_steps : ``integer``, optional
       Number of steps to simulate the process. Defaults to ``10000``.
    dt : ``float``, optional
       Process and system simulation time-step.
       Defaults to ``process.default_dt``.
    **run_steps_kwargs : ``dictionary``, optional
       Additional keyword arguments for ``process.run_steps``.

    See Also
    --------
    :class:`.EvalPoints`
    :class:`.Callable`
    :class:`.RollingWindow`
    :class:`nengo.Ensemble`
    :class:`nengo.dists.Distribution`

    Notes
    -----
    The option ``normalize_encoders=False``, required to make this useful,
    is only available in ``nengo>=2.4.0``. Otherwise these encoders will
    end up being equivalent to ``nengo.dists.Choice(np.vstack([I, -I]))``
    (i.e., unit-length and axis-aligned).
    """

    def sample(self, num, d=1, rng=np.random):
        """Samples ``n`` points in ``d`` dimensions."""
        y = self._sample(d, rng)
        r = 1./np.max(np.abs(y), axis=0)
        assert r.shape == (d,)
        rI = r*np.eye(d)
        return Choice(np.vstack([rI, -rI])).sample(num, d=d, rng=rng)



import os.path
import multiprocessing
import sys
import time

import pytest

from nengo.utils.compat import PY2
from nengo.utils.lock import FileLock, TimeoutError

if PY2:
    filter_resourcewarning = lambda fn: fn
else:
    filter_resourcewarning = pytest.mark.filterwarnings(
        'ignore::ResourceWarning')


def acquire_lock(filename):
    lock = FileLock(filename, timeout=0.01, poll=0.01)
    try:
        lock.acquire()
    except TimeoutError:
        sys.exit(0)
    else:
        sys.exit(1)


def acquire_lock_and_idle(filename):
    lock = FileLock(filename, timeout=0.01, poll=0.01)
    lock.acquire()
    while True:
        time.sleep(1.)


def test_can_acquire_filelock_at_most_once(tmpdir):
    filename = os.path.join(str(tmpdir), 'lock')
    lock = FileLock(filename, timeout=0.01, poll=0.01)
    lock.acquire()
    p = multiprocessing.Process(target=acquire_lock, args=(filename,))
    p.start()
    p.join()
    lock.release()
    assert p.exitcode == 0


@filter_resourcewarning
def test_process_termination_releases_lock(tmpdir):
    filename = os.path.join(str(tmpdir), 'lock')
    p = multiprocessing.Process(target=acquire_lock_and_idle, args=(filename,))
    p.start()
    while p.is_alive() and not os.path.exists(filename):
        time.sleep(0.2)
    assert p.is_alive()

    lock = FileLock(filename, timeout=0.01, poll=0.01)
    with pytest.raises(TimeoutError):
        lock.acquire()
    p.terminate()
    p.join()
    lock = FileLock(filename, timeout=0.01, poll=0.01)
    lock.acquire()
    lock.release()


def test_released_filelock_can_be_reacquired(tmpdir):
    filename = os.path.join(str(tmpdir), 'lock')
    lock = FileLock(filename, timeout=0.01, poll=0.01)
    lock.acquire()
    lock.release()
    lock.acquire()
    lock.release()


def test_can_release_filelock_multiple_times(tmpdir):
    filename = os.path.join(str(tmpdir), 'lock')
    lock = FileLock(filename, timeout=0.01, poll=0.01)
    lock.release()
    lock.acquire()
    lock.release()
    lock.release()


def test_filelock_supports_with_statement(tmpdir):
    filename = os.path.join(str(tmpdir), 'lock')
    with FileLock(filename):
        pass


def test_filelock_gets_released_on_lock_deletion(tmpdir):
    filename = os.path.join(str(tmpdir), 'lock')
    lock = FileLock(filename, timeout=0.01, poll=0.01)
    lock.acquire()
    del lock
    lock = FileLock(filename, timeout=0.01, poll=0.01)
    lock.acquire()
    lock.release()



#!/usr/bin/env python

import imp
import io
import os
from setuptools import setup, find_packages


def read(*filenames, **kwargs):
    encoding = kwargs.get("encoding", "utf-8")
    sep = kwargs.get("sep", "\n")
    buf = []
    for filename in filenames:
        with io.open(filename, encoding=encoding) as f:
            buf.append(f.read())
    return sep.join(buf)

name = 'nengolib'
root = os.path.dirname(os.path.realpath(__file__))
version_module = imp.load_source(
    'version', os.path.join(root, name, 'version.py'))

deps = [  # https://github.com/nengo/nengo/issues/508
    "nengo>=2.2.0,<3.0",
    "numpy>=1.13",
    "scipy>=0.19.0",
]

download_url = (
    'https://github.com/arvoelke/nengolib/archive/v%s.tar.gz' % (
        version_module.version))

setup(
    name=name,
    version=version_module.version,
    author="Aaron R. Voelker",
    author_email="arvoelke@gmail.com",
    description="Tools for robust dynamics in Nengo",
    long_description=read("README.rst", "CHANGES.rst"),
    url="https://github.com/arvoelke/nengolib/",
    download_url=download_url,
    license="Free for non-commercial use (see Nengo license)",
    packages=find_packages(),
    setup_requires=deps,
    install_requires=deps,
    keywords=[
        'Neural Engineering Framework',
        'Nengo',
        'Dynamical Spiking Networks',
        'Neural Dynamics',
        'Reservoir Computing',
    ],
    classifiers=[  # https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 5 - Production/Stable',
        'Framework :: Nengo',
        'Intended Audience :: Science/Research',
        'License :: Free for non-commercial use',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ]
)



import warnings

import numpy as np

from nengo.exceptions import ReadonlyError, SpaParseError, ValidationError
from nengo.params import Parameter
from nengo.spa import pointer
from nengo.utils.compat import is_iterable, is_number, is_integer, range


class Vocabulary(object):
    """A collection of semantic pointers, each with their own text label.

    The Vocabulary can also act as a dictionary, with keys as the names
    of the semantic pointers and values as the `.SemanticPointer` objects
    themselves. If it is asked for a pointer that does not exist, one will
    be automatically created.

    Parameters
    -----------
    dimensions : int
        Number of dimensions for each semantic pointer.
    randomize : bool, optional (Default: True)
        Whether to randomly generate pointers. If False, the semantic
        pointers will be ``[1, 0, 0, ...], [0, 1, 0, ...], [0, 0, 1, ...]``
        and so on.
    unitary : bool or list, optional (Default: False)
        If True, all generated pointers will be unitary. If a list of
        strings, any pointer whose name is in the list will be forced to be
        unitary when created.
    max_similarity : float, optional (Default: 0.1)
        When randomly generating pointers, ensure that the cosine of the
        angle between the new pointer and all existing pointers is less
        than this amount. If the system is unable to find such a pointer
        after 100 tries, a warning message is printed.
    include_pairs : bool, optional (Default: False)
        Whether to keep track of all pairs of pointers as well. This
        is helpful for determining if a vector is similar to ``A*B`` (in
        addition to being similar to ``A`` or ``B``), but exponentially
        increases the processing time.
    rng : `numpy.random.RandomState`, optional (Default: None)
        The random number generator to use to create new vectors.

    Attributes
    ----------
    include_pairs : bool
        Whether to keep track of all pairs of pointers as well. This
        is helpful for determining if a vector is similar to ``A*B`` (in
        addition to being similar to ``A`` or ``B``), but exponentially
        increases the processing time.
    key_pairs : list
        The names of all pairs of semantic pointers
        (e.g., ``['A*B', 'A*C', 'B*C']``).
    keys : list of strings
        The names of all known semantic pointers (e.g., ``['A', 'B', 'C']``).
    vector_pairs : ndarray
        The values for each pair of semantic pointers, convolved together,
        in the same order as in ``key_pairs``.
    vectors : ndarray
        All of the semantic pointer values in a matrix, in the same order
        as in ``keys``.
    """

    def __init__(self, dimensions, randomize=True, unitary=False,
                 max_similarity=0.1, include_pairs=False, rng=None):

        if not is_integer(dimensions) or dimensions < 1:
            raise ValidationError("dimensions must be a positive integer",
                                  attr='dimensions', obj=self)
        self.dimensions = dimensions
        self.randomize = randomize
        self.unitary = unitary
        self.max_similarity = max_similarity
        self.pointers = {}
        self.keys = []
        self.key_pairs = None
        self.vectors = np.zeros((0, dimensions), dtype=float)
        self.vector_pairs = None
        self._include_pairs = None
        self.include_pairs = include_pairs
        self._identity = None
        self.rng = rng
        self.readonly = False
        self.parent = None

    def create_pointer(self, attempts=100, unitary=False):
        """Create a new semantic pointer.

        This will take into account the randomize and max_similarity
        parameters from self. If a pointer satisfying max_similarity
        is not generated after the specified number of attempts, the
        candidate pointer with lowest maximum cosine with all existing
        pointers is returned.
        """
        if self.randomize:
            if self.vectors.shape[0] == 0:
                p = pointer.SemanticPointer(self.dimensions, rng=self.rng)
            else:
                p_sim = np.inf
                for _ in range(attempts):
                    pp = pointer.SemanticPointer(self.dimensions, rng=self.rng)
                    pp_sim = max(np.dot(self.vectors, pp.v))
                    if pp_sim < p_sim:
                        p = pp
                        p_sim = pp_sim
                        if p_sim < self.max_similarity:
                            break
                else:
                    warnings.warn(
                        'Could not create a semantic pointer with '
                        'max_similarity=%1.2f (D=%d, M=%d)'
                        % (self.max_similarity,
                           self.dimensions,
                           len(self.pointers)))

            # Check and make vector unitary if needed
            if unitary:
                p.make_unitary()
        else:
            index = len(self.pointers)
            if index >= self.dimensions:
                raise ValidationError(
                    "Tried to make more semantic pointers than "
                    "dimensions with non-randomized Vocabulary",
                    attr='dimensions', obj=self)
            p = pointer.SemanticPointer(np.eye(self.dimensions)[index])
        return p

    def __getitem__(self, key):
        """Return the semantic pointer with the requested name.

        If one does not exist, automatically create one. The key must be
        a valid semantic pointer name, which is any Python identifier starting
        with a capital letter.
        """
        if not key[0].isupper():
            raise SpaParseError(
                "Semantic pointers must begin with a capital letter.")
        value = self.pointers.get(key, None)
        if value is None:
            if is_iterable(self.unitary):
                unitary = key in self.unitary
            else:
                unitary = self.unitary
            value = self.create_pointer(unitary=unitary)
            self.add(key, value)
        return value

    def add(self, key, p):
        """Add a new semantic pointer to the vocabulary.

        The pointer value can be a `.SemanticPointer` or a vector.
        """
        if self.readonly:
            raise ReadonlyError(attr='Vocabulary',
                                msg="Cannot add semantic pointer '%s' to "
                                    "read-only vocabulary." % key)

        if not key[0].isupper():
            raise SpaParseError(
                "Semantic pointers must begin with a capital letter.")
        if not isinstance(p, pointer.SemanticPointer):
            p = pointer.SemanticPointer(p)

        if key in self.pointers:
            raise ValidationError("The semantic pointer %r already exists"
                                  % key, attr='pointers', obj=self)

        self.pointers[key] = p
        self.keys.append(key)
        self.vectors = np.vstack([self.vectors, p.v])

        # Generate vector pairs
        if self.include_pairs and len(self.keys) > 1:
            for k in self.keys[:-1]:
                self.key_pairs.append('%s*%s' % (k, key))
                v = (self.pointers[k] * p).v
                self.vector_pairs = np.vstack([self.vector_pairs, v])

    @property
    def include_pairs(self):
        return self._include_pairs

    @include_pairs.setter
    def include_pairs(self, value):
        """Adjusts whether key pairs are kept track of by the vocabulary.

        If this is turned on, we need to compute all the pairs of terms
        already existing.
        """
        if value == self._include_pairs:
            return
        self._include_pairs = value
        if self._include_pairs:
            self.key_pairs = []
            self.vector_pairs = np.zeros((0, self.dimensions), dtype=float)
            for i in range(1, len(self.keys)):
                for k in self.keys[:i]:
                    key = self.keys[i]
                    self.key_pairs.append('%s*%s' % (k, key))
                    v = (self.pointers[k] * self.pointers[key]).v
                    self.vector_pairs = np.vstack((self.vector_pairs, v))
        else:
            self.key_pairs = None
            self.vector_pairs = None

    def parse(self, text):
        """Evaluate a text string and return the corresponding SemanticPointer.

        This uses the Python ``eval()`` function, so any Python operators that
        have been defined for SemanticPointers are valid (``+``, ``-``, ``*``,
        ``~``, ``()``). Any terms do not exist in the vocabulary will be
        automatically generated. Valid semantic pointer terms must start
        with a capital letter.

        If the expression returns a scalar (int or float), a scaled version
        of the identity SemanticPointer will be returned.
        """

        # The following line does everything.  Note that self is being
        # passed in as the locals dictionary, and thanks to the __getitem__
        # implementation, this will automatically create new semantic
        # pointers as needed.
        try:
            value = eval(text, {}, self)
        except NameError:
            raise SpaParseError(
                "Semantic pointers must start with a capital letter.")

        if is_number(value):
            value = value * self.identity
        if not isinstance(value, pointer.SemanticPointer):
            raise SpaParseError(
                "The result of parsing '%s' is not a SemanticPointer" % text)
        return value

    @property
    def identity(self):
        """Return the identity vector."""
        if self._identity is None:
            v = np.zeros(self.dimensions)
            v[0] = 1
            self._identity = pointer.SemanticPointer(v)
        return self._identity

    def text(self, v, minimum_count=1, maximum_count=None,  # noqa: C901
             threshold=0.1, join=';', terms=None, normalize=False):
        """Return a human-readable text version of the provided vector.

        This is meant to give a quick text version of a vector for display
        purposes. To do this, compute the dot product between the vector
        and all the terms in the vocabulary. The top few vectors are
        chosen for inclusion in the text. It will try to only return
        terms with a match above the threshold, but will always return
        at least minimum_count and at most maximum_count terms. Terms
        are sorted from most to least similar.

        Parameters
        ----------
        v : SemanticPointer or ndarray
            The vector to convert into text.
        minimum_count : int, optional (Default: 1)
            Always return at least this many terms in the text.
        maximum_count : int, optional (Default: None)
            Never return more than this many terms in the text.
            If None, all terms will be returned.
        threshold : float, optional (Default: 0.1)
            How small a similarity for a term to be ignored.
        join : str, optional (Default: ';')
            The text separator to use between terms.
        terms : list, optional (Default: None)
            Only consider terms in this list of strings.
        normalize : bool, optional (Default: False)
            Whether to normalize the vector before computing similarity.
        """
        if isinstance(v, pointer.SemanticPointer):
            v = v.v
        else:
            v = np.array(v, dtype='float')

        if normalize:
            nrm = np.linalg.norm(v)
            if nrm > 0:
                v /= nrm

        m = np.dot(self.vectors, v)
        matches = [(mm, self.keys[i]) for i, mm in enumerate(m)]
        if self.include_pairs:
            m2 = np.dot(self.vector_pairs, v)
            matches2 = [(mm2, self.key_pairs[i]) for i, mm2 in enumerate(m2)]
            matches.extend(matches2)
        if terms is not None:
            # TODO: handle the terms parameter more efficiently, so we don't
            # compute a whole bunch of dot products and then throw them out
            matches = [mm for mm in matches if mm[1] in terms]
        matches.sort()
        matches.reverse()

        r = []
        for m in matches:
            if minimum_count is not None and len(r) < minimum_count:
                r.append(m)
            elif maximum_count is not None and len(r) == maximum_count:
                break
            elif threshold is None or m[0] > threshold:
                r.append(m)
            else:
                break

        return join.join(['%0.2f%s' % (sim, key) for (sim, key) in r])

    def dot(self, v):
        """Returns the dot product with all terms in the Vocabulary.

        Input parameter can either be a `.SemanticPointer` or a vector.
        """
        if isinstance(v, pointer.SemanticPointer):
            v = v.v
        return np.dot(self.vectors, v)

    def dot_pairs(self, v):
        """Returns the dot product with all pairs of terms in the Vocabulary.

        Input parameter can either be a `.SemanticPointer` or a vector.
        """
        if not self.include_pairs:
            raise ValidationError(
                "'include_pairs' must be True to call dot_pairs",
                attr='include_pairs', obj=self)

        if isinstance(v, pointer.SemanticPointer):
            v = v.v
        return np.dot(self.vector_pairs, v)

    def transform_to(self, other, keys=None):
        """Create a linear transform from one Vocabulary to another.

        This is simply the sum of the outer products of the corresponding
        terms in each Vocabulary.

        Parameters
        ----------
        other : Vocabulary
            The other vocabulary to translate into.
        keys : list, optional (Default: None)
            If None, any term that exists in just one of the Vocabularies
            will be created in the other Vocabulary and included. Otherwise,
            the transformation will only consider terms in this list. Any
            terms in this list that do not exist in the Vocabularies will
            be created.
        """
        # If the parent vocabs of self and other are the same, then no
        # transform is needed between the two vocabularies, so return an
        # identity matrix.
        my_parent = self if self.parent is None else self.parent
        other_parent = other if other.parent is None else other.parent

        if my_parent is other_parent:
            return np.eye(self.dimensions)
        else:
            if keys is None:
                if self.readonly and other.readonly:
                    keys = [k for k in self.keys if k in other.keys]
                elif self.readonly:
                    keys = list(self.keys)
                elif other.readonly:
                    keys = list(other.keys)
                else:
                    keys = list(self.keys)
                    keys.extend([k for k in other.keys if k not in self.keys])

            t = np.zeros((other.dimensions, self.dimensions), dtype=float)
            for k in keys:
                a = self[k].v
                b = other[k].v
                t += np.outer(b, a)
            return t

    def prob_cleanup(self, similarity, vocab_size, steps=10000):
        """Estimate the chance of successful cleanup.

        This returns the chance that, out of vocab_size randomly chosen
        vectors, at least one of them will be closer to a particular
        vector than the value given by compare. To use this, compare
        your noisy vector with the ideal vector, pass that value in as
        the similarity parameter, and set ``vocab_size`` to be the number of
        competing vectors.

        The steps parameter sets the accuracy of the approximate integral
        needed to compute this.

        The basic principle used here is that the probability of two random
        vectors in a D-dimensional space being a given angle apart is
        proportional to ``sin(angle)**(D-2)``.  So we integrate this value
        to get a probability of one vector being farther away than the
        desired angle, and then raise that to vocab_size to get the
        probability that all of them are farther away.
        """

        # TODO: test for numerical stability.  We are taking a number
        # slightly below 1 and raising it to a large exponent, so there's
        # lots of room for rounding errors.

        angle = np.arccos(similarity)

        x = np.linspace(0, np.pi, steps)
        y = np.sin(x) ** (self.dimensions-2)

        total = np.sum(y)
        too_close = np.sum(y[:int(angle * steps / np.pi)])

        perror1 = too_close / total

        pcorrect = (1 - perror1) ** vocab_size
        return pcorrect

    def extend(self, keys, unitary=False):
        """Extends the vocabulary with additional keys.

        Creates and adds the semantic pointers listed in keys to the
        vocabulary.

        Parameters
        ----------
        keys : list
            List of semantic pointer names to be added to the vocabulary.
        unitary : bool or list, optional (Default: False)
            If True, all generated pointers will be unitary. If a list of
            strings, any pointer whose name is on the list will be forced to
            be unitary when created.
        """
        if is_iterable(unitary):
            if is_iterable(self.unitary):
                self.unitary.extend(unitary)
            else:
                self.unitary = list(unitary)
        elif unitary:
            if is_iterable(self.unitary):
                self.unitary.extend(keys)
            else:
                self.unitary = list(keys)

        for key in keys:
            if key not in self.keys:
                self[key]

    def create_subset(self, keys):
        """Returns the subset of this vocabulary.

        Creates and returns a subset of the current vocabulary that contains
        all the semantic pointers found in keys.

        Parameters
        ----------
        keys : list
            List of semantic pointer names to be copied over to the
            new vocabulary.
        """
        # Make new Vocabulary object
        subset = Vocabulary(self.dimensions,
                            self.randomize,
                            self.unitary,
                            self.max_similarity,
                            self.include_pairs,
                            self.rng)

        # Copy over the new keys
        for key in keys:
            subset.add(key, self.pointers[key])

        # Assign the parent
        if self.parent is not None:
            subset.parent = self.parent
        else:
            subset.parent = self

        # Make the subset read only
        subset.readonly = True

        return subset


class VocabularyParam(Parameter):
    """Can be a Vocabulary."""

    def coerce(self, instance, vocab):
        self.check_type(instance, vocab, Vocabulary)
        return super(VocabularyParam, self).coerce(instance, vocab)



from __future__ import division

import numpy as np
import pytest

import nengo

from scipy.linalg import inv
from scipy.signal import lfilter

from nengolib.signal.system import (
    sys2ss, sys2zpk, sys2tf, canonical, sys_equal, ss_equal, LinearSystem,
    s, z)
from nengolib import Network, Lowpass, Alpha
from nengolib.signal import cont2discrete, shift
from nengolib.synapses import PadeDelay
from nengolib.testing import warns


def test_sys_conversions():
    sys = Alpha(0.1)

    tf = sys2tf(sys)
    ss = sys2ss(sys)
    zpk = sys2zpk(sys)

    assert sys_equal(sys2ss(tf), ss)
    assert sys_equal(sys2ss(ss), ss)  # unchanged
    assert sys_equal(sys2tf(tf), tf)  # unchanged
    assert sys_equal(sys2tf(ss), tf)

    assert sys_equal(sys2zpk(zpk), zpk)  # sanity check
    assert sys_equal(sys2zpk(tf), zpk)  # sanity check
    assert sys_equal(sys2zpk(ss), zpk)
    assert sys_equal(sys2tf(zpk), tf)
    assert sys_equal(sys2ss(zpk), ss)

    # should also work with nengo's synapse types
    assert sys_equal(sys2zpk(nengo.Alpha(0.1)), zpk)
    assert sys_equal(sys2tf(nengo.Alpha(0.1)), tf)
    assert sys_equal(sys2ss(nengo.Alpha(0.1)), ss)

    # system can also be just a scalar
    assert sys_equal(sys2tf(2.0), (1, 0.5))
    assert np.allclose(sys2ss(5)[3], 5)
    assert sys_equal(sys2zpk(5), 5)

    with pytest.raises(ValueError):
        sys2ss(np.zeros(5))

    with pytest.raises(ValueError):
        sys2zpk(np.zeros(5))

    with pytest.raises(ValueError):
        sys2tf(np.zeros(5))

    with pytest.raises(ValueError):
        # _ss2tf(...): passthrough must be single element
        sys2tf(([], [], [], [1, 2]))


def test_check_sys_equal():
    assert not sys_equal(np.ones(2), np.ones(3))

    assert s != z
    assert not z == s
    assert LinearSystem(5, analog=True) != LinearSystem(5, analog=False)

    with pytest.raises(ValueError):
        sys_equal(s, z)

    with pytest.raises(ValueError):
        ss_equal(s, z)


def test_canonical():
    sys = ([1], [1], [1], [0])
    assert sys_equal(canonical(sys), sys)

    sys = ([[1, 0], [1, 0]], [[1], [0]], [[1, 1]], [0])
    assert sys_equal(canonical(sys), sys)

    sys = ([[1, 0], [0, 1]], [[0], [1]], [[1, 1]], [0])
    assert sys_equal(canonical(sys), sys)

    sys = ([[1, 0], [0, 1]], [[1], [0]], [[1, 1]], [0])
    assert sys_equal(canonical(sys), sys)

    sys = ([[1, 0], [0, 0]], [[1], [0]], [[1, 1]], [0])
    assert sys_equal(canonical(sys), sys)

    sys = ([[1, 0], [0, 0]], [[0], [1]], [[1, 1]], [0])
    assert sys_equal(canonical(sys), sys)

    sys = ([[1, 0, 1], [0, 1, 1], [1, 0, 0]], [[0], [1], [-1]],
           [[1, 1, 1]], [0])
    assert sys_equal(canonical(sys), sys)

    sys = nengo.Alpha(0.1)
    csys = canonical(sys, controllable=True)
    osys = canonical(sys, controllable=False)

    assert ss_equal(csys, LinearSystem(sys).controllable)
    assert ss_equal(osys, LinearSystem(sys).observable)

    assert sys_equal(csys, osys)
    assert not ss_equal(csys, osys)  # different state-space realizations

    A, B, C, D = csys.ss
    assert sys_equal(csys, sys)
    assert ss_equal(csys,
                    ([[-20, -100], [1, 0]], [[1], [0]], [[0, 100]], [[0]]))

    assert sys_equal(osys, sys)
    assert ss_equal(osys,
                    ([[-20, 1], [-100, 0]], [[0], [100]], [[1, 0]], [[0]]))


def test_is_stable():
    sys = Lowpass(0.1)
    assert sys.is_stable

    assert not (~s).is_stable  # integrator

    assert LinearSystem(1).is_stable

    assert (~(z * (z - 0.5))).is_stable
    assert not (z / (z - 1)).is_stable  # discrete integrator


@pytest.mark.parametrize("sys", [PadeDelay(0.1, 4), PadeDelay(0.2, 5, 5)])
def test_decompose_states(sys):
    assert np.dot(sys.C, list(sys)) + sys.D == sys


def test_non_siso_manipulation():
    sys = Alpha(0.1)
    A, B, C, D = sys.ss

    SIMO = LinearSystem((A, B, np.eye(len(A)), [[0], [0]]))
    assert not SIMO.is_SISO
    assert SIMO.size_in == 1
    assert SIMO.size_out == 2
    assert SIMO.shape == (2, 1)
    assert not SIMO.has_passthrough
    assert ss_equal(_eval(SIMO), SIMO)
    assert isinstance(str(SIMO), str)
    assert ss_equal(canonical(SIMO), SIMO)
    for sub1, sub2 in zip(sys, SIMO):
        assert ss_equal(sub1, sub2)

    MISO = LinearSystem((A, [[1, 1]], C, [[0, 1]]))
    assert not MISO.is_SISO
    assert MISO.size_in == 2
    assert MISO.size_out == 1
    assert MISO.shape == (1, 2)
    assert MISO.has_passthrough
    assert ss_equal(_eval(MISO), MISO)
    assert isinstance(str(MISO), str)

    MIMO = LinearSystem((A, [[1, 1]], np.eye(len(A)), np.zeros((2, 2))))
    assert not MIMO.is_SISO
    assert MIMO.size_in == MIMO.size_out == 2
    assert MIMO.shape == (2, 2)
    assert not MIMO.has_passthrough
    assert ss_equal(_eval(MIMO), MIMO)
    assert isinstance(str(MIMO), str)
    for sub1, sub2 in zip(MISO, MIMO):
        assert ss_equal(sub1, sub2)


def test_non_siso_filtering(rng):
    sys = PadeDelay(0.1, order=4)
    length = 1000

    SIMO = sys.X
    assert not SIMO.is_SISO
    assert SIMO.size_in == 1
    assert SIMO.size_out == len(sys)

    x = SIMO.impulse(length)
    for i, (sub1, sub2) in enumerate(zip(sys, SIMO)):
        assert sub1 == sub2
        y1 = sub1.impulse(length)
        y2 = sub2.impulse(length)
        _transclose(shift(y1), shift(y2), x[:, i])

    B = np.asarray([[1, 2, 3], [0, 0, 0], [0, 0, 0], [0, 0, 0]]) * sys.B
    u = rng.randn(length, 3)

    Bu = u.dot([1, 2, 3])
    assert Bu.shape == (length,)
    MISO = LinearSystem((sys.A, B, sys.C, np.zeros((1, 3))), analog=True)
    assert not MISO.is_SISO
    assert MISO.size_in == 3
    assert MISO.size_out == 1

    y = cont2discrete(MISO, dt=0.001).filt(u)
    assert y.shape == (length,)
    assert np.allclose(shift(sys.filt(Bu)), y)

    MIMO = MISO.X
    assert not MIMO.is_SISO
    assert MIMO.size_in == 3
    assert MIMO.size_out == 4

    y = MIMO.filt(u)
    I = np.eye(len(sys))
    for i, sub1 in enumerate(MIMO):
        sub2 = LinearSystem((sys.A, B, I[i:i+1], np.zeros((1, 3))))
        _transclose(sub1.filt(u), sub2.filt(u), y[:, i])


def test_bad_filt():
    sys = PadeDelay(0.1, order=4).X
    with pytest.raises(ValueError):
        sys.filt(np.ones((4, 4)))
    with pytest.raises(ValueError):
        sys.filt(np.ones((4, 1)), filtfilt=True)
    with pytest.raises(ValueError):
        sys.filt(np.ones((4,)), copy=False)


@pytest.mark.parametrize("sys", [
    Lowpass(0.01), Alpha(0.2), LinearSystem(([1, 1], [0.01, 1]))])
def test_simulation(sys, Simulator, plt, seed):
    assert isinstance(sys, LinearSystem)
    old_sys = nengo.LinearFilter(sys.num, sys.den)
    assert sys == old_sys

    with Network() as model:
        stim = nengo.Node(output=nengo.processes.WhiteSignal(
            1.0, high=10, seed=seed))
        out_new = nengo.Node(size_in=2)
        out_old = nengo.Node(size_in=2)
        nengo.Connection(stim, out_new, transform=[[1], [-1]], synapse=sys)
        nengo.Connection(stim, out_old, transform=[[1], [-1]], synapse=old_sys)
        p_new = nengo.Probe(out_new)
        p_old = nengo.Probe(out_old)

    with Simulator(model) as sim:
        sim.run(1.0)

    plt.figure()
    plt.plot(sim.trange(), sim.data[p_new])
    plt.plot(sim.trange(), sim.data[p_old])

    assert np.allclose(sim.data[p_new], sim.data[p_old])


def test_discrete_synapse(Simulator):
    # Test that discrete synapses are simulated properly
    delay_steps = 50

    with Network() as model:
        stim = nengo.Node(output=np.sin)
        output = nengo.Node(size_in=1)
        nengo.Connection(stim, output, synapse=z**-delay_steps)
        p_stim = nengo.Probe(stim, synapse=None)
        p_output = nengo.Probe(output, synapse=None)

    with Simulator(model) as sim:
        sim.run(1.0)

    assert np.allclose(sim.data[p_output][delay_steps:],
                       sim.data[p_stim][:-delay_steps])


def _apply_filter(sys, dt, u):
    # "Correct" implementation of filt that has a single time-step delay
    # see Nengo issue #938
    if dt is not None:
        num, den = cont2discrete(sys, dt).tf
    elif not sys.analog:
        num, den = sys.tf
    else:
        raise ValueError("system (%s) must be discrete if not given dt" % sys)

    # convert from the polynomial representation, and add back the leading
    # zeros that were dropped by poly1d, since lfilter will shift it the
    # wrong way (it will add the leading zeros back to the end, effectively
    # removing the delay)
    num, den = map(np.asarray, (num, den))
    num = np.append([0]*(len(den) - len(num)), num)
    return lfilter(num, den, u, axis=-1)


def test_filt():
    u = np.asarray([1.0, 0, 0])
    dt = 0.1
    num, den = [1], [1, 2, 1]
    sys1 = nengo.LinearFilter(num, den)
    sys2 = LinearSystem((num, den))  # uses a different make_step
    y1 = sys1.filt(u, dt=dt, y0=0)
    y2 = sys2.filt(u, dt=dt, y0=0)
    assert np.allclose(y1, y2)


def test_filt_issue_nengo1124():
    with warns(UserWarning):
        Lowpass(0.1).filt(np.asarray([1, 0]), dt=0.001, y0=1)

    with warns(UserWarning):
        Lowpass(0.1).filt(np.asarray([1, 0]), dt=0.001, y0=None)


def test_filt_issue_64():
    # Issue #64
    # Note we need a double delay because of Nengo issue #938 (see below)
    assert np.allclose((~z**2).filt([1, 0, 0]), [0, 1, 0])


def _transclose(*arrays):
    assert len(arrays) >= 2
    for i in range(len(arrays)-1):
        assert np.allclose(arrays[i], arrays[i+1]), i


def test_filt_issue_nengo938():
    # Testing related to nengo issue #938
    # test combinations of _apply_filter / filt on nengo / nengolib
    # using a passthrough / (strictly) proper and y0=0 / y0=None
    # ... in an **ideal** world all of these would produce the same results
    # but we assert behaviour here so that it is at least documented
    # and so that we are alerted to any changes in these differences
    # https://github.com/nengo/nengo/issues/938
    # https://github.com/nengo/nengo/issues/1124

    sys_prop_nengo = nengo.LinearFilter([1], [1, 0])
    sys_prop_nglib = LinearSystem(([1], [1, 0]))
    sys_pass_nengo = nengo.LinearFilter([1e-9, 1], [1, 0])
    sys_pass_nglib = LinearSystem(([1e-9, 1], [1, 0]))

    u = np.asarray([1.0, 0.5, 0])
    dt = 0.001

    def filt_scipy(sys):
        return _apply_filter(sys, dt=dt, u=u)

    def filt_nengo(sys, y0):
        return sys.filt(u, dt=dt, y0=y0)

    # Strictly proper transfer function
    prop_nengo_apply = filt_scipy(sys_prop_nengo)
    prop_nglib_apply = filt_scipy(sys_prop_nglib)
    prop_nengo_filt0 = filt_nengo(sys_prop_nengo, y0=0)
    prop_nglib_filt0 = filt_nengo(sys_prop_nglib, y0=0)
    prop_nengo_filtN = filt_nengo(sys_prop_nengo, y0=None)
    prop_nglib_filtN = filt_nengo(sys_prop_nglib, y0=None)

    # => two equivalence classes
    _transclose(prop_nengo_apply, prop_nglib_apply)
    _transclose(prop_nengo_filt0, prop_nglib_filt0, prop_nglib_filtN)

    # One-step delay difference between these two classes
    _transclose(prop_nengo_apply[1:], prop_nengo_filt0[:-1])

    # Passthrough transfer functions
    pass_nengo_apply = filt_scipy(sys_pass_nengo)
    pass_nglib_apply = filt_scipy(sys_pass_nglib)
    pass_nengo_filt0 = filt_nengo(sys_pass_nengo, y0=0)
    pass_nglib_filt0 = filt_nengo(sys_pass_nglib, y0=0)
    pass_nengo_filtN = filt_nengo(sys_pass_nengo, y0=None)
    pass_nglib_filtN = filt_nengo(sys_pass_nglib, y0=None)

    # => almost all are equivalent (except nengo with y0=None)
    _transclose(pass_nengo_apply, pass_nglib_apply, pass_nengo_filt0,
                pass_nglib_filt0, pass_nglib_filtN)
    assert not np.allclose(prop_nengo_filtN, pass_nengo_filtN)

    # And belongs to the same equivalence class as the very first
    _transclose(prop_nengo_apply, pass_nengo_apply)


def test_sys_multiplication():
    # Check that alpha is just two lowpass multiplied together
    assert Lowpass(0.1) * Lowpass(0.1) == Alpha(0.1)


def test_sim_new_synapse(Simulator):
    # Create a new synapse object and simulate it
    synapse = Lowpass(0.1) - Lowpass(0.01)
    with Network() as model:
        stim = nengo.Node(output=np.sin)
        x = nengo.Node(size_in=1)
        nengo.Connection(stim, x, synapse=synapse)
        p_stim = nengo.Probe(stim, synapse=None)
        p_x = nengo.Probe(x, synapse=None)

    with Simulator(model) as sim:
        sim.run(0.1)

    assert np.allclose(shift(synapse.filt(sim.data[p_stim])),
                       sim.data[p_x])


def _eval(sys):
    return eval(
        repr(sys), {}, {'LinearSystem': LinearSystem, 'array': np.array})


def test_linear_system():
    tau = 0.05
    sys = Lowpass(tau)
    dsys = (1 - np.exp(-1)) * z / (1 - np.exp(-1) * z)

    # Test attributes before state-space/zpk computed
    assert sys.is_tf
    assert not sys.is_ss
    assert not sys.is_zpk

    # Test representations
    assert sys == (1, [tau, 1])
    assert sys_equal(sys.tf, sys)
    assert sys_equal(sys.ss, sys)
    assert sys_equal(sys.zpk, sys)

    # Test attributes after state-space/zpk computed
    assert sys.is_tf
    assert sys.is_ss
    assert sys.is_zpk

    # Size in/out-related properties
    assert sys.is_SISO
    assert dsys.is_SISO
    assert sys.size_in == sys.size_out == dsys.size_in == dsys.size_out == 1

    # Test attributes
    assert np.allclose(sys.num, (1/tau,))
    assert np.allclose(sys.den, (1, 1/tau))
    assert sys.causal
    assert sys.strictly_proper
    assert not sys.has_passthrough
    assert not (sys/s).has_passthrough
    assert (sys*s).has_passthrough
    assert (sys*s).causal
    assert not (sys*s).strictly_proper
    assert not (sys*s*s).has_passthrough and not (sys*s*s).causal
    assert (sys*s*s + sys*s).has_passthrough

    assert np.allclose(sys.A, -1/tau)
    assert np.allclose(sys.B, 1)
    assert np.allclose(sys.C, 1/tau)
    assert np.allclose(sys.D, 0)

    assert np.allclose(sys.zeros, [0])
    assert np.allclose(sys.poles, [-1/tau])
    assert np.allclose(sys.gain, 1/tau)
    assert np.allclose(sys.zpk[0], np.array([]))
    assert np.allclose(sys.zpk[1], np.array([-1/tau]))
    assert np.allclose(sys.zpk[2], 1/tau)

    assert sys.order_num == 0
    assert sys.order_den == 1
    assert len(sys) == 1  # order_den
    assert len(LinearSystem(sys.ss)) == 1  # uses state-space rep

    # Test dcgain and __call__
    assert np.allclose(sys.dcgain, 1)
    assert np.allclose(dsys.dcgain, 1)
    assert np.allclose((s*sys)(1e12), 1.0 / tau)  # initial value theorem
    assert np.allclose((s*sys)(0), 0)  # final value theorem
    assert np.allclose(((1 - ~z)*dsys)(1), 0)  # final value theorem

    # Test multiplication and squaring
    assert sys*2 == 2*sys
    assert (0.4*sys) + (0.6*sys) == sys
    assert sys + sys == 2*sys
    assert sys * sys == sys**2
    assert sys_equal(sys + sys*sys, sys*sys + sys)

    # Test pow
    with pytest.raises(TypeError):
        sys**0.5
    assert sys**0 == LinearSystem(1)

    # Test inversion
    inv = ~sys
    assert inv == ([tau, 1], 1)
    assert not inv.causal

    assert inv == 1 / sys
    assert inv == sys**(-1)

    # Test repr/str
    copy = _eval(sys)
    assert copy == sys

    # Test addition/subtraction
    assert sys + 2 == ((2*tau, 3), (tau, 1))
    assert 3 + sys == (-sys)*(-1) + 3
    assert (4 - sys) + 2 == (-sys) + 6
    assert np.allclose((sys - sys).num, 0)

    # Test division
    assert sys / 2 == sys * 0.5
    assert 2 / sys == 2 * inv

    cancel = sys / sys
    assert np.allclose(cancel.num, cancel.den)

    # Test inequality
    assert sys != (sys*2)

    # Test usage of differential building block
    assert sys == 1 / (tau*s + 1)


def test_linear_system_type():
    # Test that sys1 is reused by sys2
    sys1 = LinearSystem(1)
    sys2 = LinearSystem(sys1)
    sys3 = LinearSystem(1)

    assert sys1 is sys2
    assert sys1 is not sys3

    # Test that sys1 is still reused even with weird arg/kwarg ordering
    sys4 = LinearSystem(sys1, analog=True)
    sys5 = LinearSystem(sys=sys1, analog=True)
    sys6 = LinearSystem(analog=True, sys=sys1)

    assert sys1 is sys4
    assert sys1 is sys5
    assert sys1 is sys6

    # Test that analog argument gets inherited properly
    assert LinearSystem(s).analog
    assert LinearSystem(s, analog=True).analog
    assert not LinearSystem(z).analog
    assert not LinearSystem(z, analog=False).analog
    assert LinearSystem(nengo.Lowpass(0.1)).analog
    assert not LinearSystem(LinearSystem(([1], [1]), analog=False)).analog

    # Test that analog argument must match
    with pytest.raises(TypeError):
        LinearSystem(sys1, analog=False)

    with pytest.raises(TypeError):
        LinearSystem(sys1, False)

    with pytest.raises(TypeError):
        LinearSystem(sys1, analog=False)

    with pytest.raises(TypeError):
        LinearSystem(s, analog=False)

    with pytest.raises(TypeError):
        LinearSystem(z, analog=True)

    with pytest.raises(TypeError):
        LinearSystem(LinearSystem(([1], [1]), analog=True), analog=False)

    with pytest.raises(TypeError):
        LinearSystem(LinearSystem(([1], [1]), analog=False), analog=True)


def test_invalid_operations():
    with pytest.raises(ValueError):
        z + s

    with pytest.raises(ValueError):
        s - z

    with pytest.raises(ValueError):
        z * s

    with pytest.raises(ValueError):
        z / s


def test_hashing():
    assert len(set((z, s))) == 2
    assert len(set((s, 5*s/5))) == 1


def test_zerodim_system():
    sys = LinearSystem(1)
    assert len(sys) == 0
    assert ss_equal(sys, (0, 0, 0, 1))

    # However, this following system could have dimension 0 or 1
    # depending on whether we're before or after scipy 0.18
    # see https://github.com/scipy/scipy/issues/5760
    # TODO: ideally it would stay 0, but documenting this weirdness for now
    ss_sys = LinearSystem(sys.ss)
    assert len(ss_sys) in (0, 1)


def test_similarity_transform():
    sys = Alpha(0.1)

    TA, TB, TC, TD = sys.transform(np.eye(2), np.eye(2)).ss
    A, B, C, D = sys2ss(sys)
    assert np.allclose(A, TA)
    assert np.allclose(B, TB)
    assert np.allclose(C, TC)
    assert np.allclose(D, TD)

    T = [[1, 1], [-0.5, 0]]
    rsys = sys.transform(T)
    assert ss_equal(rsys, sys.transform(T, inv(T)))

    TA, TB, TC, TD = rsys.ss
    assert not np.allclose(A, TA)
    assert not np.allclose(B, TB)
    assert not np.allclose(C, TC)
    assert np.allclose(D, TD)
    assert sys_equal(sys, (TA, TB, TC, TD))

    length = 1000
    dt = 0.001
    x_old = np.asarray(
        [sub.impulse(length=length, dt=dt) for sub in sys])
    x_new = np.asarray(
        [sub.impulse(length=length, dt=dt) for sub in rsys])

    # dot(T, x_new(t)) = x_old(t)
    assert np.allclose(np.dot(T, x_new), x_old)


def test_impulse():
    dt = 0.001
    tau = 0.005
    length = 500

    delta = np.zeros(length)
    delta[0] = 1. / dt

    sys = Lowpass(tau)
    response = sys.impulse(length, dt)
    assert not np.allclose(response[0], 0)

    # should give the same result as using filt
    assert np.allclose(response, sys.filt(delta, dt))

    # and should default to the same dt
    assert sys.default_dt == dt
    assert np.allclose(response, sys.impulse(length))

    # should also accept discrete systems
    dss = cont2discrete(sys, dt=dt)
    assert not dss.analog
    assert np.allclose(response, dss.impulse(length) / dt)
    assert np.allclose(response, dss.impulse(length, dt=dt))


def test_impulse_dt():
    length = 1000
    sys = Alpha(0.1)
    # the dt should not alter the magnitude of the response
    assert np.allclose(
        max(sys.impulse(length, dt=0.001)),
        max(sys.impulse(length, dt=0.0005)),
        atol=1e-4)


def test_invalid_impulse():
    with pytest.raises(ValueError):
        s.impulse(length=10, dt=None)  # must be digital if dt is None



"""Nengo version information.

We use semantic versioning (see http://semver.org/).
and conform to PEP440 (see https://www.python.org/dev/peps/pep-0440/).
'.devN' will be added to the version unless the code base represents
a release version. Release versions are git tagged with the version.
"""

name = "nengo"
version_info = (2, 8, 0)  # (major, minor, patch)
dev = None

version = "{v}{dev}".format(v='.'.join(str(v) for v in version_info),
                            dev=('.dev%d' % dev) if dev is not None else '')



import numpy as np

from nengo.builder import Builder, Signal
from nengo.builder.operator import Reset, SimPyFunc
from nengo.exceptions import BuildError
from nengo.node import Node
from nengo.processes import Process
from nengo.utils.compat import is_array_like


@Builder.register(Node)
def build_node(model, node):
    """Builds a `.Node` object into a model.

    The node build function is relatively simple. It involves creating input
    and output signals, and connecting them with an `.Operator` that depends
    on the type of ``node.output``.

    Parameters
    ----------
    model : Model
        The model to build into.
    node : Node
        The node to build.

    Notes
    -----
    Sets ``model.params[node]`` to ``None``.
    """

    # input signal
    if not is_array_like(node.output) and node.size_in > 0:
        sig_in = Signal(np.zeros(node.size_in), name="%s.in" % node)
        model.add_op(Reset(sig_in))
    else:
        sig_in = None

    # Provide output
    if node.output is None:
        sig_out = sig_in
    elif isinstance(node.output, Process):
        sig_out = Signal(np.zeros(node.size_out), name="%s.out" % node)
        model.build(node.output, sig_in, sig_out)
    elif callable(node.output):
        sig_out = (Signal(np.zeros(node.size_out), name="%s.out" % node)
                   if node.size_out > 0 else None)
        model.add_op(SimPyFunc(
            output=sig_out, fn=node.output, t=model.time, x=sig_in))
    elif is_array_like(node.output):
        sig_out = Signal(node.output, name="%s.out" % node)
    else:
        raise BuildError(
            "Invalid node output type %r" % type(node.output).__name__)

    model.sig[node]['in'] = sig_in
    model.sig[node]['out'] = sig_out
    model.params[node] = None



import nengo
from nengo.spa.module import Module
from nengo.utils.compat import iteritems


def make_parse_func(func, vocab):
    """Create a function that calls func and parses the output in vocab."""

    def parse_func(t):
        return vocab.parse(func(t)).v

    return parse_func


class Input(Module):
    """A SPA module for providing external inputs to other modules.

    The parameters passed to this module indicate the module input name
    and the function to execute to generate inputs to that module.
    The functions should always return strings, which will then be parsed
    by the relevant vocabulary. For example::

        def input1(t):
            if t < 0.1:
                return 'A'
            else:
                return '0'

        spa_net.input = spa.Input(vision=input1, task='X')

    will create two inputs:

    1. an input to the ``vision`` module, which for the first 0.1 seconds
       is the value associated with the ``'A'`` semantic pointer and then
       a vector of all zeros, and
    2. an input to the ``task`` module which is always the value associated
       with the ``'X'`` semantic pointer.

    Parameters
    ----------
    label : str, optional (Default: None)
        A name for the ensemble. Used for debugging and visualization.
    seed : int, optional (Default: None)
        The seed used for random number generation.
    add_to_container : bool, optional (Default: None)
        Determines if this Network will be added to the current container.
        If None, will be true if currently within a Network.
    """

    def __init__(self, label=None, seed=None, add_to_container=None, **kwargs):
        super(Input, self).__init__(label, seed, add_to_container)
        self.kwargs = kwargs
        self.input_nodes = {}

    def on_add(self, spa):
        """Create the connections and nodes."""
        Module.on_add(self, spa)

        for name, value in iteritems(self.kwargs):
            target, vocab = spa.get_module_input(name)
            if callable(value):
                val = make_parse_func(value, vocab)
            else:
                val = vocab.parse(value).v

            with self:
                node = nengo.Node(val, label='input_%s' % name)
            self.input_nodes[name] = node

            with spa:
                nengo.Connection(node, target, synapse=None)



import numpy as np

from nengolib.signal.utils import nrmse, shift
from nengolib.stats import sphere


def test_nrmse():
    assert np.allclose(nrmse(np.ones(100), np.ones(100)), 0)
    assert np.allclose(nrmse([0, 0, 0], [42, 9000, 0]), 1)

    ideal = sphere.sample(100, 3)
    r = np.asarray([1, 2, 3])
    assert np.allclose(nrmse(ideal, r*ideal, axis=0), 1-1./r)


def test_shift():
    a = [1, 2, 3]
    assert np.allclose(shift(a), [0, 1, 2])

    assert np.allclose(shift(a, 2, 42), [42, 42, 1])

    assert np.allclose(shift(a, 1000), [0, 0, 0])

    a = [[1, 2], [3, 4], [5, 6]]
    assert np.allclose(shift(a), [[0, 0], [1, 2], [3, 4]])
    assert np.allclose(shift(a, 2, [0, 1]), [[0, 1], [0, 1], [1, 2]])
    assert np.allclose(shift(a, 2, 1), [[1, 1], [1, 1], [1, 2]])



import os
import sys


if sys.platform.startswith('win'):
    config_dir = os.path.expanduser(os.path.join("~", ".nengo"))
    cache_dir = os.path.join(config_dir, "cache")
else:
    config_dir = os.path.expanduser(os.path.join("~", ".config", "nengo"))
    cache_dir = os.path.expanduser(os.path.join("~", ".cache", "nengo"))

decoder_cache_dir = os.path.join(cache_dir, "decoders")
install_dir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), os.pardir, os.pardir))
examples_dir = os.path.join(install_dir, "docs", "examples")

nengorc = {'system': os.path.join(install_dir, 'nengo-data', 'nengorc'),
           'user': os.path.join(config_dir, "nengorc"),
           'project': os.path.abspath(os.path.join(os.curdir, "nengorc"))}



import numpy as np

from nengo.base import Process
from nengo.exceptions import ValidationError
from nengo.utils.stdlib import checked_call

__all__ = ['Callable']


class Callable(Process):
    """Adapter to convert any callable into a :class:`nengo.Process`.

    Parameters
    ----------
    func : ``callable``
        A function that can be called with a single float argument (time).
    default_dt : ``float``, optional
        Default time-step for the process. Defaults to ``0.001`` seconds.
    seed : ``integer``, optional
        Seed for the process.

    See Also
    --------
    :class:`nengo.Process`

    Examples
    --------
    Making a sine wave process using a lambda:

    >>> from nengolib.processes import Callable
    >>> process = Callable(lambda t: np.sin(2*np.pi*t))

    >>> import matplotlib.pyplot as plt
    >>> plt.plot(process.ntrange(1000), process.run_steps(1000))
    >>> plt.xlabel("Time (s)")
    >>> plt.show()
    """

    def __init__(self, func, default_dt=0.001, seed=None):
        if not callable(func):
            raise ValidationError("func (%s) must be callable" % func,
                                  attr='func', obj=self)
        self.func = func
        value, invoked = checked_call(func, 0.)
        if not invoked:
            raise ValidationError(
                "func (%s) must only take single float argument" % func,
                attr='func', obj=self)
        default_size_out = np.asarray(value).size
        super(Callable, self).__init__(
            default_size_in=0,
            default_size_out=default_size_out,
            default_dt=default_dt,
            seed=seed)

    def __repr__(self):
        return "%s(func=%r, default_dt=%r, seed=%r)" % (
            type(self).__name__, self.func, self.default_dt, self.seed)

    def make_step(self, shape_in, shape_out, dt, rng):
        if shape_in != (0,):
            raise ValidationError("shape_in must be (0,), got %s" % (
                shape_in,), attr='func', obj=self)
        if shape_out != (self.default_size_out,):
            raise ValidationError("shape_out must be (%s,), got %s" % (
                self.default_size_out, shape_out), attr='func', obj=self)

        def step(t):
            return self.func(t)
        return step



from nengo.utils.cache import byte_align, bytes2human, human2bytes


def test_bytes2human():
    assert bytes2human(1) == '1.0 B'
    assert bytes2human(10000) == '9.8 KB'
    assert bytes2human(100001221) == '95.4 MB'


def test_human2bytes():
    assert human2bytes('1 MB') == 1048576
    assert human2bytes('1.5 GB') == 1610612736
    assert human2bytes('14 B') == 14
    assert human2bytes('1B') == 1
    assert human2bytes('1   B') == 1


def test_byte_align():
    assert byte_align(5, 16) == 16
    assert byte_align(23, 8) == 24
    assert byte_align(13, 1) == 13
    assert byte_align(0, 16) == 0
    assert byte_align(32, 8) == 32



from __future__ import absolute_import

import sys

import numpy as np
import pytest

import nengo
import nengo.utils.numpy as npext
from nengo.base import Process
from nengo.dists import Distribution, Gaussian
from nengo.exceptions import ValidationError
from nengo.processes import (BrownNoise, FilteredNoise, Piecewise,
                             PresentInput, WhiteNoise, WhiteSignal)
from nengo.synapses import Lowpass


class DistributionMock(Distribution):
    def __init__(self, retval):
        super(DistributionMock, self).__init__()
        self.retval = retval
        self.sample_calls = []

    def sample(self, n, d, rng=np.random):
        self.sample_calls.append((n, d, rng))
        return np.ones((n, d)) * self.retval


class TimeProcess(Process):
    def make_step(self, shape_in, shape_out, dt, rng):
        size_in = np.prod(shape_in)
        size_out = np.prod(shape_out)
        if size_in == 0:
            return lambda t: [t] * size_out
        else:
            return lambda t, x: [t * np.sum(x)] * size_out


def test_time(RefSimulator):
    trun = 1.
    c = 2.
    process = TimeProcess()

    with nengo.Network() as model:
        u = nengo.Node(process)
        up = nengo.Probe(u)

        q = nengo.Node(c)
        v = nengo.Node(process, size_in=1, size_out=1)
        nengo.Connection(q, v, synapse=None)
        vp = nengo.Probe(v)

    with RefSimulator(model) as sim:
        sim.run(trun)

    nt = len(sim.trange())
    assert process.trange(trun).shape == sim.trange().shape
    assert np.allclose(process.trange(trun), sim.trange())
    assert np.allclose(process.ntrange(nt), sim.trange())
    assert np.allclose(process.run(trun), sim.data[up])
    assert np.allclose(process.apply([c] * nt), sim.data[vp])


def test_whitenoise(rng):
    dist = DistributionMock(42)
    process = WhiteNoise(dist, scale=False)
    samples = process.run_steps(5, d=2, rng=rng)
    assert np.all(samples == 42 * np.ones((5, 2)))
    assert process.run_steps(5, rng=rng).shape == (5, 1)
    assert process.run_steps(1, d=1, rng=rng).shape == (1, 1)
    assert process.run_steps(2, d=3, rng=rng).shape == (2, 3)


def test_brownnoise(Simulator, seed, plt):
    d = 5000
    t = 0.5
    std = 1.5
    process = BrownNoise(dist=Gaussian(0, std))
    with nengo.Network() as model:
        u = nengo.Node(process, size_out=d)
        up = nengo.Probe(u)

    with Simulator(model, seed=seed) as sim:
        sim.run(t)
    samples = sim.data[up]

    trange = sim.trange()
    expected_std = std * np.sqrt(trange)
    atol = 3.5 * expected_std / np.sqrt(d)

    plt.subplot(2, 1, 1)
    plt.title("Five Brown noise signals")
    plt.plot(trange, samples[:, :5])
    plt.subplot(2, 1, 2)
    plt.ylabel("Standard deviation")
    plt.plot(trange, np.abs(np.std(samples, axis=1)), label="Actual")
    plt.plot(trange, expected_std, label="Expected")
    plt.legend(loc='best')

    assert np.all(np.abs(np.mean(samples, axis=1)) < atol)
    assert np.all(np.abs(np.std(samples, axis=1) - expected_std) < atol)


def psd(values, dt=0.001):
    freq = npext.rfftfreq(values.shape[0], d=dt)
    power = 2. * np.std(np.abs(np.fft.rfft(
        values, axis=0)), axis=1) / np.sqrt(values.shape[0])
    return freq, power


@pytest.mark.parametrize('rms', [0.5, 1, 100])
def test_gaussian_whitenoise(Simulator, rms, seed, plt):
    d = 500
    process = WhiteNoise(Gaussian(0., rms), scale=False)
    with nengo.Network() as model:
        u = nengo.Node(process, size_out=d)
        up = nengo.Probe(u)

    with Simulator(model, seed=seed) as sim:
        sim.run(0.3)
    values = sim.data[up]
    freq, val_psd = psd(values, dt=sim.dt)

    trange = sim.trange()
    plt.subplot(2, 1, 1)
    plt.title("First two dimensions of white noise process, rms=%.1f" % rms)
    plt.plot(trange, values[:, :2])
    plt.xlim(right=trange[-1])
    plt.subplot(2, 1, 2)
    plt.title("Power spectrum")
    plt.plot(freq, val_psd, drawstyle='steps')

    val_rms = npext.rms(values, axis=0)
    assert np.allclose(val_rms.mean(), rms, rtol=0.02)
    assert np.allclose(val_psd[1:-1], rms, rtol=0.2)


@pytest.mark.parametrize('rms', [0.5, 1, 100])
def test_whitesignal_rms(Simulator, rms, seed, plt):
    t = 1.
    d = 500
    process = WhiteSignal(t, high=500, rms=rms)
    with nengo.Network() as model:
        u = nengo.Node(process, size_out=d)
        up = nengo.Probe(u)

    with Simulator(model, seed=seed) as sim:
        sim.run(t)
    values = sim.data[up]
    freq, val_psd = psd(values, dt=sim.dt)

    trange = sim.trange()
    plt.subplot(2, 1, 1)
    plt.title("First two D of white noise process, rms=%.1f" % rms)
    plt.plot(trange, values[:, :2])
    plt.xlim(right=trange[-1])
    plt.subplot(2, 1, 2)
    plt.title("Power spectrum")
    plt.plot(freq, val_psd, drawstyle='steps')

    assert np.allclose(np.std(values), rms, rtol=0.02)
    assert np.allclose(val_psd[1:-1], rms, rtol=0.35)


@pytest.mark.parametrize('y0,d', [(0, 1), (-0.3, 3), (0.4, 1)])
def test_whitesignal_y0(Simulator, seed, y0, d):
    t = .1
    process = WhiteSignal(t, high=500, y0=y0)
    with nengo.Network() as model:
        u = nengo.Node(process, size_out=d)
        up = nengo.Probe(u)

    with Simulator(model, seed=seed) as sim:
        sim.run(t)
    values = sim.data[up]
    error = np.min(abs(y0 - values), axis=0)

    assert ((y0 - error <= values[0, :]) & (values[0, :] <= y0 + error)).all()


@pytest.mark.parametrize('high,dt', [(10, 0.01), (5, 0.001), (50, 0.001)])
def test_whitesignal_high_dt(Simulator, high, dt, seed, plt):
    t = 1.
    rms = 0.5
    d = 500
    process = WhiteSignal(t, high, rms=rms)
    with nengo.Network() as model:
        u = nengo.Node(process, size_out=d)
        up = nengo.Probe(u)

    with Simulator(model, seed=seed, dt=dt) as sim:
        sim.run(t)
    values = sim.data[up]
    freq, val_psd = psd(values, dt=dt)

    trange = sim.trange()
    plt.subplot(2, 1, 1)
    plt.title("First two D of white noise process, high=%d Hz" % high)
    plt.plot(trange, values[:, :2])
    plt.xlim(right=trange[-1])
    plt.subplot(2, 1, 2)
    plt.title("Power spectrum")
    plt.plot(freq, val_psd, drawstyle='steps')
    plt.xlim(0, high * 2.0)

    assert np.allclose(np.std(values, axis=1), rms, rtol=0.15)
    assert np.all(val_psd[npext.rfftfreq(len(trange), dt) > high] < rms * 0.5)


@pytest.mark.parametrize('high,dt', [(501, 0.001), (500, 0.002)])
def test_whitesignal_nyquist(Simulator, dt, high, seed):
    # check that high cannot exceed nyquist frequency
    process = WhiteSignal(1.0, high=high)
    with nengo.Network() as model:
        nengo.Node(process, size_out=1)

    with pytest.raises(ValidationError):
        Simulator(model, dt=dt, seed=seed)


def test_whitesignal_continuity(Simulator, seed, plt):
    """Test that WhiteSignal is continuous over multiple periods."""
    t = 1.
    high = 10
    rms = 0.5
    process = WhiteSignal(t, high=high, rms=rms)
    with nengo.Network() as model:
        u = nengo.Node(process, size_out=1)
        up = nengo.Probe(u)

    with Simulator(model, seed=seed) as sim:
        sim.run(4 * t)
    dt = sim.dt
    x = sim.data[up]

    plt.plot(sim.trange(), x)

    # tolerances approximated from derivatives of sine wave of highest freq
    safety_factor = 2.
    a, f = np.sqrt(2) * rms, (2 * np.pi * high) * dt
    assert abs(np.diff(x, axis=0)).max() <= safety_factor * a * f
    assert abs(np.diff(x, n=2, axis=0)).max() <= safety_factor**2 * a * f**2


def test_sampling_shape():
    process = WhiteSignal(0.1, high=500)
    assert process.run_steps(1).shape == (1, 1)
    assert process.run_steps(5, d=1).shape == (5, 1)
    assert process.run_steps(1, d=2). shape == (1, 2)


def test_reset(Simulator, seed):
    trun = 0.1

    with nengo.Network() as model:
        u = nengo.Node(WhiteNoise(Gaussian(0, 1), scale=False))
        up = nengo.Probe(u)

    with Simulator(model, seed=seed) as sim:
        sim.run(trun)
        x = np.array(sim.data[up])
        sim.reset()
        sim.run(trun)
        y = np.array(sim.data[up])

    assert x.shape == y.shape
    assert np.allclose(x, y)


def test_frozen():
    """Test attributes inherited from FrozenObject"""
    a = WhiteNoise(dist=Gaussian(0.3, 0.2))
    b = WhiteNoise(dist=Gaussian(0.3, 0.2))
    c = FilteredNoise(dist=Gaussian(0.3, 0.2), synapse=Lowpass(0.02))

    assert hash(a) == hash(a)
    assert hash(b) == hash(b)
    assert hash(c) == hash(c)

    assert a == b
    assert hash(a) == hash(b)
    assert a != c
    assert hash(a) != hash(c)  # not guaranteed, but highly likely
    assert b != c
    assert hash(b) != hash(c)  # not guaranteed, but highly likely

    with pytest.raises(ValueError):
        a.dist = Gaussian(0.3, 0.5)  # test that dist param is frozen
    with pytest.raises(ValueError):
        a.dist.std = 0.4  # test that dist object is frozen


def test_seed(Simulator, seed):
    with nengo.Network() as model:
        a = nengo.Node(WhiteSignal(0.1, high=100, seed=seed))
        b = nengo.Node(WhiteSignal(0.1, high=100, seed=seed+1))
        c = nengo.Node(WhiteSignal(0.1, high=100))
        d = nengo.Node(WhiteNoise(seed=seed))
        e = nengo.Node(WhiteNoise())
        ap = nengo.Probe(a)
        bp = nengo.Probe(b)
        cp = nengo.Probe(c)
        dp = nengo.Probe(d)
        ep = nengo.Probe(e)

    with Simulator(model) as sim1:
        sim1.run(0.1)

    with Simulator(model) as sim2:
        sim2.run(0.1)

    tols = dict(atol=1e-7, rtol=1e-4)
    assert np.allclose(sim1.data[ap], sim2.data[ap], **tols)
    assert np.allclose(sim1.data[bp], sim2.data[bp], **tols)
    assert not np.allclose(sim1.data[cp], sim2.data[cp], **tols)
    assert not np.allclose(sim1.data[ap], sim1.data[bp], **tols)
    assert np.allclose(sim1.data[dp], sim2.data[dp], **tols)
    assert not np.allclose(sim1.data[ep], sim2.data[ep], **tols)


def test_present_input(Simulator, rng):
    n = 5
    c, ni, nj = 3, 8, 8
    images = rng.normal(size=(n, c, ni, nj))
    pres_time = 0.1

    model = nengo.Network()
    with model:
        u = nengo.Node(PresentInput(images, pres_time))
        up = nengo.Probe(u)

    with Simulator(model) as sim:
        sim.run(1.0)

    t = sim.trange()
    i = (np.floor((t - sim.dt) / pres_time + 1e-7) % n).astype(int)
    y = sim.data[up].reshape(len(t), c, ni, nj)
    for k, [ii, image] in enumerate(zip(i, y)):
        assert np.allclose(image, images[ii], rtol=1e-4, atol=1e-7), (k, ii)


class TestPiecewise(object):

    def run_sim(self, data, interpolation, Simulator):
        process = Piecewise(data, interpolation=interpolation)

        with nengo.Network() as model:
            u = nengo.Node(process, size_out=process.default_size_out)
            up = nengo.Probe(u)

        with Simulator(model) as sim:
            sim.run(0.15)

        return sim.trange(), sim.data[up]

    def test_basic(self, Simulator):
        t, f = self.run_sim({0.05: 1, 0.1: 0}, 'zero', Simulator)
        assert np.allclose(f[t == 0.001], [0.])
        assert np.allclose(f[t == 0.025], [0.])
        assert np.allclose(f[t == 0.049], [0.])
        assert np.allclose(f[t == 0.05], [1.])
        assert np.allclose(f[t == 0.075], [1.])
        assert np.allclose(f[t == 0.1], [0.])
        assert np.allclose(f[t == 0.15], [0.])

    def test_lists(self, Simulator):
        t, f = self.run_sim({0.05: [1, 0], 0.1: [0, 1]}, 'zero', Simulator)
        assert np.allclose(f[t == 0.001], [0., 0.])
        assert np.allclose(f[t == 0.025], [0., 0.])
        assert np.allclose(f[t == 0.049], [0., 0.])
        assert np.allclose(f[t == 0.05], [1., 0.])
        assert np.allclose(f[t == 0.075], [1., 0.])
        assert np.allclose(f[t == 0.1], [0., 1.])
        assert np.allclose(f[t == 0.15], [0., 1.])

    def test_default_zero(self):
        process = Piecewise({0.05: 1, 0.1: 0})
        f = process.make_step(shape_in=(process.default_size_in,),
                              shape_out=(process.default_size_out,),
                              dt=process.default_dt,
                              rng=None)
        assert np.allclose(f(-10), [0.])
        assert np.allclose(f(0), [0.])

    def test_invalid_key(self):
        data = {0.05: 1, 0.1: 0, 'a': 0.2}
        with pytest.raises(ValidationError):
            Piecewise(data)

    def test_invalid_length(self):
        data = {0.05: [1, 0], 0.1: [1, 0, 0]}
        with pytest.raises(ValidationError):
            Piecewise(data)

        # check that scalars and length 1 arrays can be used interchangeably
        # (no validation error)
        Piecewise({0.05: [1], 0.1: 0})

    def test_invalid_interpolation_type(self):
        data = {0.05: 1, 0.1: 0}
        with pytest.raises(ValidationError):
            Piecewise(data, interpolation='not-interpolation')

    def test_fallback_to_zero(self, Simulator, monkeypatch):
        # Emulate not having scipy in case we have scipy
        monkeypatch.setitem(sys.modules, "scipy.interpolate", None)

        with pytest.warns(UserWarning):
            process = Piecewise({0.05: 1, 0.1: 0}, interpolation='linear')
        assert process.interpolation == 'zero'

    def test_interpolation_1d(self, plt, Simulator):
        pytest.importorskip('scipy')

        # Note: cubic requires an explicit start of 0
        data = {0: -0.5, 0.05: 1, 0.075: -1, 0.1: -0.5}

        def test_and_plot(interp):
            t, f = self.run_sim(data, interp, Simulator)
            assert np.allclose(f[t == 0.05], [1.])
            assert np.allclose(f[t == 0.075], [-1.])
            assert np.allclose(f[t == 0.1], [-0.5])
            plt.plot(t, f, label=interp)

        test_and_plot('zero')
        test_and_plot('linear')
        test_and_plot('nearest')
        test_and_plot('slinear')
        test_and_plot('quadratic')
        test_and_plot('cubic')
        plt.legend(loc="lower left")

    def test_interpolation_2d(self, plt, Simulator):
        pytest.importorskip('scipy')

        # Note: cubic requires an explicit start of 0
        data = {
            0: [-0.5, -0.5],
            0.05: [1.0, 0.5],
            0.075: [-1, -0.5],
            0.1: [-0.5, -0.25]
        }

        ax1 = plt.subplot(2, 1, 1)
        ax2 = plt.subplot(2, 1, 2)

        def test_and_plot(interp):
            t, f = self.run_sim(data, interp, Simulator)
            assert np.allclose(f[t == 0.05], [1., 0.5])
            assert np.allclose(f[t == 0.075], [-1., -0.5])
            assert np.allclose(f[t == 0.1], [-0.5, -0.25])
            ax1.plot(t, f.T[0], label=interp)
            ax2.plot(t, f.T[1], label=interp)

        test_and_plot('zero')
        test_and_plot('linear')
        test_and_plot('nearest')
        test_and_plot('slinear')
        test_and_plot('quadratic')
        test_and_plot('cubic')
        plt.legend(loc="lower left")

    @pytest.mark.parametrize("value", (1, [1, 1]))
    def test_shape_out(self, value):
        process = Piecewise({1: value})
        f = process.make_step(shape_in=(process.default_size_in,),
                              shape_out=(process.default_size_out,),
                              dt=process.default_dt,
                              rng=None)

        assert np.array_equal(f(0), np.zeros(process.default_size_out))
        assert np.array_equal(f(2), np.ones(process.default_size_out))

    def test_function(self, Simulator):
        t, f = self.run_sim({0.05: np.sin, 0.1: np.cos}, 'zero', Simulator)
        assert np.allclose(f[t == 0.001], [0.])
        assert np.allclose(f[t == 0.049], [0.])
        assert np.allclose(f[t == 0.05], [np.sin(0.05)])
        assert np.allclose(f[t == 0.075], [np.sin(0.075)])
        assert np.allclose(f[t == 0.1], [np.cos(0.1)])
        assert np.allclose(f[t == 0.15], [np.cos(0.15)])

    def test_function_list(self, Simulator):
        def func1(t):
            return t, t**2, t**3

        def func2(t):
            return t**4, t**5, t**6

        t, f = self.run_sim({0.05: func1, 0.1: func2}, 'zero', Simulator)
        assert np.allclose(f[t == 0.001], [0.])
        assert np.allclose(f[t == 0.049], [0.])
        assert np.allclose(f[t == 0.05], func1(0.05))
        assert np.allclose(f[t == 0.075], func1(0.075))
        assert np.allclose(f[t == 0.1], func2(0.1))
        assert np.allclose(f[t == 0.15], func2(0.15))

    def test_mixture(self, Simulator):
        t, f = self.run_sim({0.05: 1, 0.1: np.cos}, 'zero', Simulator)
        assert np.allclose(f[t == 0.001], [0.])
        assert np.allclose(f[t == 0.049], [0.])
        assert np.allclose(f[t == 0.05], [1.])
        assert np.allclose(f[t == 0.075], [1.])
        assert np.allclose(f[t == 0.1], [np.cos(0.1)])
        assert np.allclose(f[t == 0.15], [np.cos(0.15)])

    def test_mixture_3d(self, Simulator):
        def func(t):
            return t, t**2, t**3

        t, f = self.run_sim({0.05: [1, 1, 1], 0.1: func}, 'zero', Simulator)
        assert np.allclose(f[t == 0.001], [0., 0., 0.])
        assert np.allclose(f[t == 0.049], [0., 0., 0.])
        assert np.allclose(f[t == 0.05], [1., 1., 1.])
        assert np.allclose(f[t == 0.075], [1., 1., 1.])
        assert np.allclose(f[t == 0.1], func(0.1))
        assert np.allclose(f[t == 0.15], func(0.15))

    def test_invalid_function_length(self, Simulator):
        with pytest.raises(ValidationError):
            Piecewise({0.5: 0, 1.0: lambda t: [t, t ** 2]})

    def test_invalid_interpolation_on_func(self):
        def func(t):
            return t

        with pytest.warns(UserWarning):
            process = Piecewise({0.05: 0, 0.1: func}, interpolation='linear')
        assert process.interpolation == 'zero'



import nengo
from nengo import Connection as BaseConnection
from nengo.solvers import LstsqL2

from nengolib.solvers import BiasedSolver

__all__ = ['Connection']


class Connection(BaseConnection):
    """Drop-in replacement for :class:`nengo.Connection`.

    Extends :class:`nengo.Connection` to improve decoding by using a bias.
    If the ``pre`` object is a :class:`nengo.Ensemble`, then a bias will
    be included in its decoders. This automatically improves the decoding of
    functions with a constant offset when the ensemble has few neurons.
    This is equivalent to a change in postsynaptic biases (in effect
    changing the representation to be centered around some constant offset,
    that is discovered optimally).

    Parameters
    ----------
    pre : :class:`nengo.Ensemble` or :class:`nengo.ensemble.Neurons` or :class:`nengo.Node`
        Nengo source object for the connection.
    post : :class:`nengo.Ensemble` or :class:`nengo.ensemble.Neurons` or :class:`nengo.Node` or :class:`nengo.Probe`
        Nengo destination object for the connection.
    solver : :class:`nengo.solvers.Solver`, optional
        Solver to use for decoded bias and connection.
        Defaults to :class:`nengo.solvers.LstsqL2`.
    **kwargs : ``dictionary``, optional
        Additional keyword arguments passed to :class:`nengo.Connection`.

    See Also
    --------
    :class:`nengo.Connection`
    :class:`.Network`
    """  # noqa: E501

    def __init__(self, pre, post, solver=LstsqL2(), **kwargs):
        if isinstance(pre, nengo.Ensemble):
            solver = BiasedSolver(solver)
            bias = nengo.Node(output=solver.bias_function(post.size_in),
                              label="Bias")
            nengo.Connection(bias, post, synapse=None)

        super(Connection, self).__init__(pre, post, solver=solver, **kwargs)



import numpy as np
import pytest

from nengo.dists import Uniform, UniformHypersphere
from nengo.utils.numpy import norm

from nengolib.stats.ntmdists import (
    SphericalCoords, Sobol, Rd,
    ScatteredCube, ScatteredHypersphere, cube, sphere, ball)
from nengolib.testing import warns


@pytest.mark.parametrize("m", [1, 2, 4, 16])
def test_spherical_coords(m, rng):
    dist = SphericalCoords(m)
    n = 10001
    x = np.linspace(0, 1, n)
    pdf = dist.pdf(x)
    cdf = dist.cdf(x)

    # Check that the cdf matches the pdf
    assert np.allclose(pdf.cumsum()/n, cdf, atol=1e-3)

    # Check that the ppf inverts the cdf
    # cutoff needed because the ppf rounds up to 1 when the cdf values
    # get too close 1o 1
    assert np.allclose(dist.ppf(cdf)[x < 0.9], x[x < 0.9])

    # Check that the sampling approximates the cdf
    rvs = dist.sample(n, rng=rng)
    width = 0.1
    for b in np.arange(0, 1, 1.0/width):
        assert np.allclose(np.sum((rvs > b) & (rvs <= b+width)) / float(n),
                           dist.cdf(b+width) - dist.cdf(b), atol=1e-2)


@pytest.mark.parametrize("ntm", [Sobol(), Rd()])
def test_ntm_properties(ntm):
    # check derministic
    s1 = ntm.sample(3, 4)
    s2 = ntm.sample(3, 4)
    assert np.allclose(s1, s2)

    # check shape
    assert s1.shape == (3, 4)


@pytest.mark.parametrize("ntm", [Sobol(), Rd()])
def test_ntm_invalid_dims(ntm):
    with pytest.raises(ValueError):
        ntm.sample(1, d=0)

    with pytest.raises(ValueError):
        ntm.sample(1, d=1.5)

    if isinstance(ntm, Sobol):
        with warns(UserWarning):
            ntm.sample(2, d=41)


def _furthest(x):
    # returns largest distance from each point
    diffs = x[:, None, :] - x[None, :, :]
    dists = norm(diffs, axis=2)
    return np.max(dists, axis=0)


def _compare_samples(x1, x2, num_moments=3, atol=0.1):
    # compare each raw sample moment within some absolute tolerance
    for i in range(num_moments):
        m1 = np.mean(x1**(i + 1), axis=0)
        m2 = np.mean(x2**(i + 1), axis=0)
        assert np.allclose(m1, m2, atol=atol), i


def test_cube_bounds(rng):
    n = 100
    x = ScatteredCube(-2, 3).sample(n, 4, rng)
    assert (x >= -2).all()
    assert (x < -1.5).any()
    assert (x <= 3).all()
    assert (x > 2.5).any()

    x = ScatteredCube([-2, -1], [3, 4]).sample(n, 2, rng)
    assert (x[:, 0] >= -2).all()
    assert (x[:, 1] >= -1).all()
    assert (x[:, 0] <= 3).all()
    assert (x[:, 1] <= 4).all()

    # check that inverted bounds still work sensibly
    x = ScatteredCube([3, 4], [-2, -1]).sample(n, 2, rng)
    assert (x[:, 0] >= -2).all()
    assert (x[:, 1] >= -1).all()
    assert (x[:, 0] <= 3).all()
    assert (x[:, 1] <= 4).all()


@pytest.mark.parametrize("d", [1, 2, 4, 16, 64])
@pytest.mark.parametrize("ntm", [Sobol(), Rd()])
def test_cube(d, ntm, rng):
    cube = ScatteredCube(base=ntm)

    n = 1000
    x = cube.sample(n, d, rng)
    assert x.shape == (n, d)
    assert abs(np.mean(x)) <= 1e-2

    _compare_samples(x, Uniform(-1, 1).sample(n, d, rng))

    low = np.min(x, axis=0)
    high = np.max(x, axis=0)
    assert low.shape == high.shape == (d,)

    assert (x >= -1).all()
    assert (x <= 1).all()

    assert (low <= -0.96).all(), np.max(low)
    assert (high >= 0.96).all(), np.min(high)


@pytest.mark.parametrize("d", [1, 2, 4, 16, 64])
@pytest.mark.parametrize("ntm", [Sobol(), Rd()])
def test_ball(d, ntm, rng):
    ball = ScatteredHypersphere(surface=False, base=ntm)

    n = 1000
    x = ball.sample(n, d, rng)
    assert x.shape == (n, d)
    assert abs(np.mean(x)) < 0.1 / d

    _compare_samples(x, UniformHypersphere(surface=False).sample(n, d, rng))

    dist = norm(x, axis=1)
    assert (dist <= 1).all()

    f = _furthest(x)
    assert (f > dist + 0.5).all()


@pytest.mark.parametrize("d", [1, 2, 4, 16, 64])
@pytest.mark.parametrize("ntm", [Sobol(), Rd()])
def test_sphere(d, ntm, rng):
    sphere = ScatteredHypersphere(surface=True, base=ntm)

    n = 1000
    x = sphere.sample(n, d, rng)
    assert x.shape == (n, d)
    assert abs(np.mean(x)) < 0.1 / d

    _compare_samples(x, UniformHypersphere(surface=True).sample(n, d, rng))

    assert np.allclose(norm(x, axis=1), 1)

    f = _furthest(x)
    assert (f > 1.5).all()


def test_dist_repr():
    assert repr(SphericalCoords(4)) == "SphericalCoords(4)"
    assert repr(Sobol()) == "Sobol()"
    assert repr(Rd()) == "Rd()"
    assert (repr(cube) ==
            "ScatteredCube(low=array([-1]), high=array([1]), base=Rd())")
    assert repr(sphere) == "ScatteredHypersphere(surface=True, base=Rd())"
    assert repr(ball) == "ScatteredHypersphere(surface=False, base=Rd())"



from __future__ import division

import numpy as np

import nengo.utils.numpy as npext
from nengo.exceptions import SignalError
from nengo.utils.compat import StringIO, is_integer


class Signal(object):
    """Represents data or views onto data within a Nengo simulation.

    Signals are tightly coupled to NumPy arrays, which is how live data is
    represented in a Nengo simulation. Signals provide a view onto the
    important metadata of the live NumPy array, and maintain the original
    value of the array in order to reset the simulation to the initial state.

    Parameters
    ----------
    initial_value : array_like
        The initial value of the signal. Much of the metadata tracked by the
        Signal is based on this array as well (e.g., dtype).
    name : str, optional (Default: None)
        Name of the signal. Primarily used for debugging.
        If None, the memory location of the Signal will be used.
    base : Signal, optional (Default: None)
        The base signal, if this signal is a view on another signal.
        Linking the two signals with the ``base`` argument is necessary
        to ensure that their live data is also linked.
    readonly : bool, optional (Default: False)
        Whether this signal and its related live data should be marked as
        readonly. Writing to these arrays will raise an exception.
    offset : int, optional (Default: 0)
        For a signal view this gives the offset of the view from the base
        ``initial_value`` in bytes. This might differ from the offset
        of the NumPy array view provided as ``initial_value`` if the base
        is a view already (in which case the signal base offset will be 0
        because it starts where the view starts. That NumPy view can have
        an offset of itself).
    """

    # Set assert_named_signals True to raise an Exception
    # if model.signal is used to create a signal with no name.
    # This can help to identify code that's creating un-named signals,
    # if you are trying to track down mystery signals that are showing
    # up in a model.
    assert_named_signals = False

    def __init__(self, initial_value,
                 name=None, base=None, readonly=False, offset=0):
        if self.assert_named_signals:
            assert name
        self._name = name

        if not np.isscalar(initial_value) and base is None:
            self._initial_value = np.ascontiguousarray(initial_value).view()
        else:
            self._initial_value = np.asarray(initial_value).view()
        self._initial_value.setflags(write=False)

        if base is not None:
            assert isinstance(base, Signal) and not base.is_view
            # make sure initial_value uses the same data as base.initial_value
            assert initial_value.base is base.initial_value.base
        self._base = base
        self._offset = offset

        self._readonly = bool(readonly)

    def __getstate__(self):
        state = dict(self.__dict__)
        v = self._initial_value
        state['_initial_value'] = (
            v.shape, v.base, npext.array_offset(v), v.strides)
        return state

    def __setstate__(self, state):
        for k, v in state.items():
            setattr(self, k, v)

        shape, base, offset, strides = self._initial_value
        self._initial_value = np.ndarray(
            shape, buffer=base, offset=offset, strides=strides)
        self._initial_value.setflags(write=False)

    def __getitem__(self, item):
        """Index or slice into array"""
        if item is Ellipsis or (
                isinstance(item, slice) and item == slice(None)):
            return self

        if not isinstance(item, tuple):
            item = (item,)

        if not all(is_integer(i) or isinstance(i, slice) for i in item):
            raise SignalError("Can only index or slice into signals")

        if all(map(is_integer, item)):
            # turn one index into slice to get a view from numpy
            item = item[:-1] + (slice(item[-1], item[-1]+1),)

        view = self._initial_value[item]
        offset = (npext.array_offset(view)
                  - npext.array_offset(self._initial_value))
        return Signal(view, name="%s[%s]" % (self.name, item),
                      base=self.base, offset=offset)

    def __repr__(self):
        return "Signal(%s, shape=%s)" % (self._name, self.shape)

    @property
    def base(self):
        """(Signal or None) The base signal, if this signal is a view.

        Linking the two signals with the ``base`` argument is necessary
        to ensure that their live data is also linked.
        """
        return self if self._base is None else self._base

    @property
    def dtype(self):
        """(numpy.dtype) Data type of the signal (e.g., float64)."""
        return self.initial_value.dtype

    @property
    def elemoffset(self):
        """(int) Offset of data from base in elements."""
        return self.offset // self.itemsize

    @property
    def elemstrides(self):
        """(int) Strides of data in elements."""
        return tuple(s // self.itemsize for s in self.strides)

    @property
    def initial_value(self):
        """(numpy.ndarray) Initial value of the signal.

        Much of the metadata tracked by the Signal is based on this array
        as well (e.g., dtype).
        """
        return self._initial_value

    @initial_value.setter
    def initial_value(self, val):
        raise SignalError("Cannot change initial value after initialization")

    @property
    def is_view(self):
        """(bool) True if this Signal is a view on another Signal."""
        return self._base is not None

    @property
    def itemsize(self):
        """(int) Size of an array element in bytes."""
        return self.initial_value.itemsize

    @property
    def name(self):
        """(str) Name of the signal. Primarily used for debugging."""
        return self._name if self._name is not None else ("0x%x" % id(self))

    @name.setter
    def name(self, name):
        self._name = name

    @property
    def nbytes(self):
        """(int) Number of bytes consumed by the signal."""
        # Equivalent to self.itemsize * self.size
        return self.initial_value.nbytes

    @property
    def ndim(self):
        """(int) Number of array dimensions."""
        return self.initial_value.ndim

    @property
    def offset(self):
        """(int) Offset of data from base in bytes.

        For a signal view this gives the offset of the view from the base
        ``initial_value`` in bytes. This might differ from the offset
        of the NumPy array view provided as ``initial_value`` if the base
        is a view already (in which case the signal base offset will be 0
        because it starts where the view starts. That NumPy view can have
        an offset of itself).
        """
        return self._offset

    @property
    def readonly(self):
        """(bool) Whether associated live data can be changed."""
        return self._readonly

    @readonly.setter
    def readonly(self, readonly):
        self._readonly = bool(readonly)

    @property
    def shape(self):
        """(tuple) Tuple of array dimensions."""
        return self.initial_value.shape

    @property
    def size(self):
        """(int) Total number of elements."""
        return self.initial_value.size

    @property
    def strides(self):
        """(tuple) Strides of data in bytes."""
        return self.initial_value.strides

    def column(self):
        """Return a view on this signal with column vector shape."""
        return self.reshape((self.size, 1))

    def may_share_memory(self, other):
        """Determine if two signals might overlap in memory.

        This comparison is not exact and errs on the side of false positives.
        See `numpy.may_share_memory` for more details.

        Parameters
        ----------
        other : Signal
            The other signal we are investigating.
        """
        return np.may_share_memory(self.initial_value, other.initial_value)

    def reshape(self, *shape):
        """Return a view on this signal with a different shape.

        Note that ``reshape`` cannot change the overall size of the signal.
        See `numpy.reshape` for more details.

        Any number of integers can be passed to this method,
        describing the desired shape of the returned signal.
        """
        return Signal(self._initial_value.reshape(*shape),
                      name="%s.reshape(%s)" % (self.name, shape),
                      base=self.base)

    def row(self):
        """Return a view on this signal with row vector shape."""
        return self.reshape((1, self.size))


class SignalDict(dict):
    """Map from Signal -> ndarray

    This dict subclass ensures that the ndarray values aren't overwritten,
    and instead data are written into them, which ensures that
    these arrays never get copied, which wastes time and space.

    Use ``init`` to set the ndarray initially.
    """
    def __getitem__(self, key):
        try:
            return dict.__getitem__(self, key)
        except KeyError:
            if isinstance(key, Signal) and key.base is not key:
                # return a view on the base signal
                base = dict.__getitem__(self, key.base)
                return np.ndarray(
                    buffer=base, dtype=key.dtype, shape=key.shape,
                    offset=key.offset, strides=key.strides)
            else:
                raise

    def __setitem__(self, key, val):
        """Ensures that ndarrays stay in the same place in memory.

        Unlike normal dicts, this means that you cannot add a new key
        to a SignalDict using __setitem__. This is by design, to avoid
        silent typos when debugging Simulator. Every key must instead
        be explicitly initialized with SignalDict.init.
        """
        self[key][...] = val

    def __str__(self):
        """Pretty-print the signals and current values."""
        sio = StringIO()
        for k in self:
            sio.write("%s %s\n" % (repr(k), repr(self[k])))
        return sio.getvalue()

    def init(self, signal):
        """Set up a permanent mapping from signal -> ndarray."""
        if signal in self:
            raise SignalError("Cannot add signal twice")

        x = signal.initial_value
        if signal.is_view:
            if signal.base not in self:
                self.init(signal.base)

            # get a view onto the base data
            view = np.ndarray(
                shape=x.shape, strides=x.strides, offset=signal.offset,
                dtype=x.dtype, buffer=self[signal.base].data)
            assert np.array_equal(view, x)
            view.setflags(write=not signal.readonly)
            dict.__setitem__(self, signal, view)
        else:
            x = x.view() if signal.readonly else x.copy()
            dict.__setitem__(self, signal, x)

    def reset(self, signal):
        """Reset ndarray to the base value of the signal that maps to it"""
        if not signal.readonly:
            self[signal] = signal.initial_value



import numpy as np
import pytest

import nengo
from nengo import spa


def test_basic():
    with spa.SPA() as model:
        model.state = spa.State(dimensions=16)

    input = model.get_module_input('state')
    output = model.get_module_output('state')
    assert input[0] is model.state.input
    assert output[0] is model.state.output
    assert input[1] is output[1]
    assert input[1].dimensions == 16


def test_neurons():
    with spa.SPA() as model:
        model.state = spa.State(dimensions=16, neurons_per_dimension=2)

    assert len(model.state.state_ensembles.ensembles) == 1
    assert model.state.state_ensembles.ensembles[0].n_neurons == 16 * 2

    with spa.SPA() as model:
        model.state = spa.State(dimensions=16, subdimensions=1,
                                neurons_per_dimension=2)

    assert len(model.state.state_ensembles.ensembles) == 16
    assert model.state.state_ensembles.ensembles[0].n_neurons == 2


def test_dimension_exception():
    with pytest.raises(Exception):
        with spa.SPA() as model:
            vocab = spa.Vocabulary(16)
            model.state = spa.State(dimensions=12, vocab=vocab)

    with spa.SPA() as model:
        model.state = spa.State(dimensions=12, subdimensions=3)


def test_no_feedback_run(Simulator, seed):
    with spa.SPA(seed=seed) as model:
        model.state = spa.State(dimensions=32, feedback=0.0)

        def state_input(t):
            if 0 <= t < 0.2:
                return 'A'
            elif 0.2 <= t < 0.4:
                return 'B'
            else:
                return '0'
        model.state_input = spa.Input(state=state_input)

    state, vocab = model.get_module_output('state')

    with model:
        p = nengo.Probe(state, 'output', synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.5)

    data = np.dot(sim.data[p], vocab.vectors.T)
    assert data[200, 0] > 0.9
    assert data[200, 1] < 0.2
    assert data[400, 0] < 0.2
    assert data[400, 1] > 0.9
    assert data[499, 0] < 0.2
    assert data[499, 1] < 0.2


def test_memory_run(Simulator, seed, plt):
    with spa.SPA(seed=seed) as model:
        model.memory = spa.State(dimensions=32, feedback=1.0,
                                 feedback_synapse=0.01)

        def state_input(t):
            if 0 <= t < 0.05:
                return 'A'
            else:
                return '0'

        model.state_input = spa.Input(memory=state_input)

    memory, vocab = model.get_module_output('memory')

    with model:
        p = nengo.Probe(memory, 'output', synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.5)
    t = sim.trange()

    similarity = np.dot(sim.data[p], vocab.vectors.T)
    plt.plot(t, similarity)
    plt.ylabel("Similarity to 'A'")
    plt.xlabel("Time (s)")

    # value should peak above 1.0, then decay down to near 1.0
    assert np.mean(similarity[(t > 0.05) & (t < 0.1)]) > 1.2
    assert np.mean(similarity[(t > 0.2) & (t < 0.3)]) > 0.8
    assert np.mean(similarity[t > 0.49]) > 0.7


def test_memory_run_decay(Simulator, plt, seed):
    with spa.SPA(seed=seed) as model:
        model.memory = spa.State(dimensions=32, feedback=(1.0 - 0.01/0.05),
                                 feedback_synapse=0.01)

        def state_input(t):
            if 0 <= t < 0.05:
                return 'A'
            else:
                return '0'

        model.state_input = spa.Input(memory=state_input)

    memory, vocab = model.get_module_output('memory')

    with model:
        p = nengo.Probe(memory, 'output', synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.3)
    data = np.dot(sim.data[p], vocab.vectors.T)

    t = sim.trange()
    plt.plot(t, data)

    assert data[t == 0.05, 0] > 1.0
    assert data[t == 0.299, 0] < 0.4



import numpy as np
import pytest

import nengo
from nengo import spa


def test_basic():
    with spa.SPA() as model:
        model.buffer = spa.Buffer(dimensions=16)

    input = model.get_module_input('buffer')
    output = model.get_module_output('buffer')
    assert input[0] is model.buffer.state.input
    assert output[0] is model.buffer.state.output
    assert input[1] is output[1]
    assert input[1].dimensions == 16


def test_neurons():
    with spa.SPA() as model:
        model.buffer = spa.Buffer(dimensions=16, neurons_per_dimension=2)

    assert len(model.buffer.state.ensembles) == 1
    assert model.buffer.state.ensembles[0].n_neurons == 16 * 2

    with spa.SPA() as model:
        model.buffer = spa.Buffer(dimensions=16, subdimensions=1,
                                  neurons_per_dimension=2)

    assert len(model.buffer.state.ensembles) == 16
    assert model.buffer.state.ensembles[0].n_neurons == 2


def test_exception():
    with pytest.raises(Exception):
        with spa.SPA() as model:
            vocab = spa.Vocabulary(16)
            model.buffer = spa.Buffer(dimensions=12, vocab=vocab)

    with spa.SPA() as model:
        model.buffer = spa.Buffer(dimensions=12, subdimensions=3)


def test_run(Simulator, seed):
    with spa.SPA(seed=seed) as model:
        model.buffer = spa.Buffer(dimensions=32)

        def input(t):
            if 0 <= t < 0.2:
                return 'A'
            elif 0.2 <= t < 0.4:
                return 'B'
            else:
                return '0'
        model.input = spa.Input(buffer=input)

    buffer, vocab = model.get_module_output('buffer')

    with model:
        p = nengo.Probe(buffer, 'output', synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.5)

    data = np.dot(sim.data[p], vocab.vectors.T)
    assert data[200, 0] > 0.9
    assert data[200, 1] < 0.2
    assert data[400, 0] < 0.2
    assert data[400, 1] > 0.9
    assert data[499, 0] < 0.2
    assert data[499, 1] < 0.2
